#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

from pyomo.environ import (
    ConcreteModel,
    Objective,
    Expression,
    Constraint,
    TransformationFactory,
    value,
    Param,
)
from pyomo.environ import units as pyunits
from pyomo.network import Arc
from idaes.core import FlowsheetBlock
from idaes.core.util.scaling import calculate_scaling_factors
from idaes.core.util.initialization import propagate_state
from watertap.examples.flowsheets.full_treatment_train.flowsheet_components import (
    pretreatment_NF,
    desalination,
    translator_block,
    gypsum_saturation_index,
    costing,
)
from watertap.examples.flowsheets.full_treatment_train.model_components import (
    property_models,
)
from watertap.examples.flowsheets.full_treatment_train.util import (
    solve_block,
    check_dof,
)

"""Flowsheet example that satisfy minimum viable product requirements"""


def build_flowsheet_mvp_NF(m, **kwargs):
    """
    Build a flowsheet with NF pretreatment and RO.
    """
    # set up keyword arguments for the sections of treatment train
    kwargs_pretreatment = {
        k: kwargs[k]
        for k in (
            "has_bypass",
            "NF_type",
            "NF_base",
        )
    }
    kwargs_desalination = {
        k: kwargs[k]
        for k in (
            "has_desal_feed",
            "is_twostage",
            "has_ERD",
            "RO_type",
            "RO_base",
            "RO_level",
        )
    }
    # build flowsheet
    property_models.build_prop(m, base="ion")
    pretrt_port = pretreatment_NF.build_pretreatment_NF(m, **kwargs_pretreatment)

    property_models.build_prop(m, base=kwargs["RO_base"])
    desal_port = desalination.build_desalination(m, **kwargs_desalination)

    property_models.build_prop(m, base="eNRTL")

    translator_block.build_tb(
        m,
        base_inlet=kwargs["NF_base"],
        base_outlet=kwargs["RO_base"],
        name_str="tb_pretrt_to_desal",
    )

    # set up Arcs between pretreatment and desalination
    m.fs.s_pretrt_tb = Arc(
        source=pretrt_port["out"], destination=m.fs.tb_pretrt_to_desal.inlet
    )
    m.fs.s_tb_desal = Arc(
        source=m.fs.tb_pretrt_to_desal.outlet, destination=desal_port["in"]
    )

    # add gypsum saturation index calculations
    gypsum_saturation_index.build(m, section="desalination", **kwargs_desalination)
    gypsum_saturation_index.build(m, section="pretreatment", **kwargs_desalination)

    # new initialization
    if kwargs["NF_type"] == "ZO":
        m.fs.NF.area.fix(175)
    if kwargs["has_bypass"]:
        m.fs.splitter.split_fraction[0, "bypass"].fix(0.50)
    m.fs.RO.area.fix(80)
    m.fs.pump_RO.control_volume.properties_out[0].pressure.fix(60e5)
    if kwargs["is_twostage"]:
        m.fs.RO2.area.fix(20)
        m.fs.pump_RO2.control_volume.properties_out[0].pressure.fix(90e5)

    # touch some properties used in optimization
    if kwargs["is_twostage"]:
        product_water_sb = m.fs.mixer_permeate.mixed_state[0]
        RO_waste_sb = m.fs.RO2.feed_side.properties[0, 1]
    else:
        product_water_sb = m.fs.RO.mixed_permeate[0]
        RO_waste_sb = m.fs.RO.feed_side.properties[0, 1]

    # NOTE: Building the costing here means it gets
    #       initialized during the simulation phase.
    #       This helps model stability.
    m.fs.feed.properties[0].flow_vol
    m.fs.feed.properties[0].conc_mol_phase_comp["Liq", "Ca"]

    m.fs.tb_pretrt_to_desal.properties_in[0].flow_vol
    m.fs.tb_pretrt_to_desal.properties_in[0].conc_mol_phase_comp["Liq", "Ca"]

    product_water_sb.flow_vol
    RO_waste_sb.flow_vol

    m.fs.system_recovery = Expression(
        expr=product_water_sb.flow_vol / m.fs.feed.properties[0].flow_vol
    )
    m.fs.total_work = Expression(
        expr=m.fs.pump_RO.work_mechanical[0]
        + (m.fs.pump_RO2.work_mechanical[0] if kwargs["is_twostage"] else 0.0)
    )

    # annual water production
    m.fs.treated_flow_vol = Expression(expr=product_water_sb.flow_vol)
    costing.build_costing(m, **kwargs)

    return m


def set_up_optimization(m, system_recovery=0.7, **kwargs_flowsheet):
    is_twostage = kwargs_flowsheet["is_twostage"]

    # scale
    calculate_scaling_factors(m)

    # unfix variables
    m.fs.splitter.split_fraction[0, "bypass"].unfix()
    m.fs.splitter.split_fraction[0, "bypass"].setlb(0.001)
    m.fs.splitter.split_fraction[0, "bypass"].setub(0.99)

    m.fs.NF.area.unfix()
    m.fs.NF.area.setlb(10)
    m.fs.NF.area.setub(1000)

    m.fs.pump_RO.control_volume.properties_out[0].pressure.unfix()
    m.fs.pump_RO.control_volume.properties_out[0].pressure.setlb(20e5)
    m.fs.pump_RO.control_volume.properties_out[0].pressure.setub(75e5)

    m.fs.RO.area.unfix()
    m.fs.RO.area.setlb(10)
    m.fs.RO.area.setub(300)

    # Set lower bound for water flux at the RO outlet, based on a minimum net driving pressure, NDPmin
    m.fs.RO.NDPmin = Param(initialize=1e5, mutable=True, units=pyunits.Pa)

    m.fs.RO.flux_mass_phase_comp[0, 1, "Liq", "H2O"].setlb(
        value(m.fs.RO.A_comp[0, "H2O"] * m.fs.RO.dens_solvent * m.fs.RO.NDPmin)
    )

    if is_twostage:
        m.fs.max_allowable_pressure = Param(
            initialize=120e5, mutable=True, units=pyunits.pascal
        )
        m.fs.pump_RO2.control_volume.properties_out[0].pressure.unfix()
        m.fs.pump_RO2.control_volume.properties_out[0].pressure.setlb(20e5)
        m.fs.pump_RO2.control_volume.properties_out[0].pressure.setub(
            m.fs.max_allowable_pressure
        )

        m.fs.RO2.area.unfix()
        m.fs.RO2.area.setlb(10)
        m.fs.RO2.area.setub(300)

        # Set lower bound for water flux at the RO outlet, based on a minimum net driving pressure, NDPmin
        m.fs.RO2.NDPmin = Param(initialize=1e5, mutable=True, units=pyunits.Pa)
        m.fs.RO2.flux_mass_phase_comp[0, 1, "Liq", "H2O"].setlb(
            value(m.fs.RO2.A_comp[0, "H2O"] * m.fs.RO2.dens_solvent * m.fs.RO2.NDPmin)
        )

    # add additional constraints
    # fixed system recovery
    m.fs.system_recovery_target = Param(initialize=system_recovery, mutable=True)
    m.fs.system_recovery_tol = Param(initialize=5e-3, mutable=True)
    m.fs.eq_system_recovery = Constraint(
        expr=(
            m.fs.system_recovery_target,
            m.fs.system_recovery,
            m.fs.system_recovery_target + m.fs.system_recovery_tol,
        )
    )

    # saturation index
    m.fs.max_saturation_index = Param(initialize=1.0, mutable=True)
    m.fs.eq_max_saturation_index_desal = Constraint(
        expr=m.fs.desal_saturation.saturation_index <= m.fs.max_saturation_index
    )

    m.fs.max_conc_factor_target = Param(initialize=3.5, mutable=True)
    m.fs.eq_max_conc_NF = Constraint(
        expr=m.fs.NF.feed_side.properties_out[0].mass_frac_phase_comp["Liq", "Ca"]
        <= m.fs.max_conc_factor_target
        * m.fs.feed.properties[0].mass_frac_phase_comp["Liq", "Ca"]
    )

    # set objective
    m.fs.objective = Objective(expr=m.fs.costing.LCOW)

    # set additional constraints to limit local minima
    # NOTE: doesn't seem necessary with new objective
    if False:
        m.fs.inequality_RO_area = Constraint(expr=m.fs.RO.area >= m.fs.RO2.area)
        min_pressure_increase = 1e5
        m.fs.inequality_RO_pressure = Constraint(
            expr=m.fs.pump_RO.control_volume.properties_out[0].pressure
            + min_pressure_increase
            <= m.fs.pump_RO2.control_volume.properties_out[0].pressure
        )
        m.fs.inequality_RO_permeate = Constraint(
            expr=m.fs.RO.permeate_side.properties_mixed[0].flow_vol_phase["Liq"]
            >= m.fs.RO2.permeate_side.properties_mixed[0].flow_vol_phase["Liq"]
        )

    check_dof(m, dof_expected=6 if is_twostage else 4)
    # solve_block(m, tee=False, fail_flag=True)


def optimize(m):
    solve_block(m, tee=False, fail_flag=True)


def solve_flowsheet_mvp_NF(**kwargs):
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    build_flowsheet_mvp_NF(m, **kwargs)
    TransformationFactory("network.expand_arcs").apply_to(m)

    # scale
    pretreatment_NF.scale_pretreatment_NF(m, **kwargs)
    calculate_scaling_factors(m.fs.tb_pretrt_to_desal)
    desalination.scale_desalination(m, **kwargs)
    calculate_scaling_factors(m)

    # initialize
    optarg = {"nlp_scaling_method": "user-scaling"}
    pretreatment_NF.initialize_pretreatment_NF(m, **kwargs)
    m.fs.pretrt_saturation.properties.initialize(optarg=optarg)
    propagate_state(m.fs.s_pretrt_tb)
    m.fs.tb_pretrt_to_desal.initialize(optarg=optarg)
    propagate_state(m.fs.s_tb_desal)
    desalination.initialize_desalination(m, **kwargs)
    m.fs.desal_saturation.properties.initialize()

    m.fs.costing.initialize()

    # check_build(m)
    # check_scaling(m)

    check_dof(m)
    solve_block(m, tee=False, fail_flag=True)

    pretreatment_NF.display_pretreatment_NF(m, **kwargs)
    m.fs.tb_pretrt_to_desal.report()
    desalination.display_desalination(m, **kwargs)
    print(
        "desalination solubility index:", value(m.fs.desal_saturation.saturation_index)
    )
    print(
        "pretreatment solubility index:", value(m.fs.pretrt_saturation.saturation_index)
    )
    print("water recovery:", value(m.fs.system_recovery))
    print("LCOW:", value(m.fs.costing.LCOW))
    print("CP modulus:", value(m.fs.desal_saturation.cp_modulus))

    return m


def solve_optimization(system_recovery=0.75, **kwargs_flowsheet):

    m = solve_flowsheet_mvp_NF(**kwargs_flowsheet)

    print("\n****** Optimization *****\n")
    set_up_optimization(m, system_recovery=system_recovery, **kwargs_flowsheet)
    optimize(m)

    pretreatment_NF.display_pretreatment_NF(m, **kwargs_flowsheet)
    m.fs.tb_pretrt_to_desal.report()
    desalination.display_desalination(m, **kwargs_flowsheet)
    costing.display_costing(m)
    print(
        "desalination saturation index:", value(m.fs.desal_saturation.saturation_index)
    )
    print(
        "pretreatment saturation index:", value(m.fs.pretrt_saturation.saturation_index)
    )
    print(
        "pretreatment Ca concentration factor:",
        value(
            m.fs.NF.feed_side.properties_out[0].mass_frac_phase_comp["Liq", "Ca"]
            / m.fs.feed.properties[0].mass_frac_phase_comp["Liq", "Ca"]
        ),
    )
    print("water recovery:", value(m.fs.system_recovery))
    print("CP modulus:", value(m.fs.desal_saturation.cp_modulus))
    return m


if __name__ == "__main__":
    import sys

    kwargs_flowsheet = {
        "has_bypass": True,
        "has_desal_feed": False,
        "is_twostage": True,
        "has_ERD": True,
        "NF_type": "ZO",
        "NF_base": "ion",
        "RO_type": "0D",
        "RO_base": "TDS",
        "RO_level": "detailed",
    }
    if len(sys.argv) == 1:
        m = solve_flowsheet_mvp_NF(**kwargs_flowsheet)
    else:
        m = solve_optimization(system_recovery=float(sys.argv[1]), **kwargs_flowsheet)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pytest
import os
import copy
import numpy as np
import pyomo.environ as pyo

from watertap.tools.parameter_sweep.sampling_types import (
    NormalSample,
    GeomSample,
    UniformSample,
)
from watertap.tools.parameter_sweep import (
    DifferentialParameterSweep,
    differential_parameter_sweep,
)
from watertap.tools.parameter_sweep.tests.test_parameter_sweep import (
    _read_output_h5,
    _get_rank0_path,
    _assert_dictionary_correctness,
    _assert_h5_csv_agreement,
    _optimization,
    _reinitialize,
    build_model,
    build_sweep_params_for_tps,
)
import watertap.tools.MPI as MPI

from watertap.tools.parameter_sweep.model_manager import ModelManager


def build_none_outputs(model):
    return None


@pytest.fixture
def model():
    m = pyo.ConcreteModel()
    m.fs = fs = pyo.Block()

    fs.input = pyo.Var(["a", "b"], within=pyo.UnitInterval, initialize=0.5)
    fs.output = pyo.Var(["c", "d"], within=pyo.UnitInterval, initialize=0.5)

    fs.slack = pyo.Var(["ab_slack", "cd_slack"], bounds=(0, 0), initialize=0.0)
    fs.slack_penalty = pyo.Param(default=1000.0, mutable=True, within=pyo.PositiveReals)

    fs.ab_constr = pyo.Constraint(
        expr=(fs.output["c"] + fs.slack["ab_slack"] == 2 * fs.input["a"])
    )
    fs.cd_constr = pyo.Constraint(
        expr=(fs.output["d"] + fs.slack["cd_slack"] == 3 * fs.input["b"])
    )

    fs.performance = pyo.Expression(expr=pyo.summation(fs.output))

    m.objective = pyo.Objective(
        expr=m.fs.performance - m.fs.slack_penalty * pyo.summation(m.fs.slack),
        sense=pyo.maximize,
    )
    return m


@pytest.mark.component
def test_check_differential_sweep_key_validity(model):
    m = model
    A = m.fs.input["a"]
    B = m.fs.input["b"]
    differential_sweep_specs = {
        A.name: {
            "diff_mode": "sum",
            "diff_sample_type": NormalSample,
            "std_dev": 0.01,
            "pyomo_object": m.fs.input["a"],
        },
        B.name: {
            "diff_mode": "product",
            "diff_sample_type": UniformSample,
            "relative_lb": 0.01,
            "relative_ub": 0.01,
            "pyomo_object": m.fs.input["b"],
        },
    }
    build_spec = lambda model: differential_sweep_specs
    sweep_params = {A.name: (A, 0.1, 0.9, 3), B.name: (B, 0.0, 0.5, 3)}

    ps = DifferentialParameterSweep(build_differential_sweep_specs=build_spec)

    sweep_params, _ = ps._process_sweep_params(sweep_params)
    ps.outputs = None
    ps._check_differential_sweep_key_validity(differential_sweep_specs, sweep_params)

    assert ps.diff_spec_index == [0, 1]


@pytest.mark.component
def test_create_differential_sweep_params_normal(model):
    m = model
    differential_sweep_specs = {
        "fs.a": {
            "diff_sample_type": NormalSample,
            "std_dev": 0.01,
            "pyomo_object": m.fs.input["a"],
        },
        "fs.b": {
            "diff_sample_type": NormalSample,
            "std_dev": 0.5,
            "pyomo_object": m.fs.input["b"],
        },
    }
    build_spec = lambda model: differential_sweep_specs
    ps = DifferentialParameterSweep(build_differential_sweep_specs=build_spec)
    ps.model_manager = ModelManager(ps)
    ps.model_manager.model = m

    local_values = np.array([0.0, 0.0, 1.0, 2.0])

    ps.diff_spec_index = [0, 1]
    diff_sweep_param_dict = ps._create_differential_sweep_params(local_values)

    expected_dict = {
        "fs.a": NormalSample(m.fs.input["a"], 0.0, 0.01, 1),
        "fs.b": NormalSample(m.fs.input["b"], 1.0, 0.5, 1),
    }

    for key, value in diff_sweep_param_dict.items():
        assert value.mean == expected_dict[key].mean
        assert value.sd == expected_dict[key].sd


@pytest.mark.component
def test_create_differential_sweep_params_sum_prod(model):
    m = model
    differential_sweep_specs = {
        "fs.a": {
            "diff_mode": "sum",
            "diff_sample_type": GeomSample,
            "relative_lb": 0.01,
            "relative_ub": 10.0,
            "pyomo_object": m.fs.input["a"],
        },
        "fs.b": {
            "diff_mode": "product",
            "diff_sample_type": UniformSample,
            "relative_lb": 0.01,
            "relative_ub": 0.1,
            "pyomo_object": m.fs.input["b"],
        },
    }
    build_spec = lambda model: differential_sweep_specs
    ps = DifferentialParameterSweep(build_differential_sweep_specs=build_spec)
    ps.model_manager = ModelManager(ps)
    ps.model_manager.model = m
    local_values = np.array([0.0, 0.1, 1.0, 2.0])

    ps.diff_spec_index = [0, 1]
    diff_sweep_param_dict = ps._create_differential_sweep_params(local_values)

    expected_dict = {
        "fs.a": GeomSample(m.fs.input["a"], 0.099, 1.1, 1),
        "fs.b": UniformSample(m.fs.input["b"], 0.01, 0.1, 1),
    }

    for key, value in diff_sweep_param_dict.items():
        assert value.lower_limit == expected_dict[key].lower_limit
        assert value.upper_limit == expected_dict[key].upper_limit


@pytest.mark.component
def test_create_differential_sweep_params_percentile(model):
    m = model
    differential_sweep_specs = {
        "fs.b": {
            "diff_mode": "percentile",
            "diff_sample_type": UniformSample,
            "relative_lb": 0.01,
            "relative_ub": 0.1,
            "nominal_lb": 0.0,
            "nominal_ub": 1.0,
            "pyomo_object": m.fs.input["b"],
        },
    }
    build_spec = lambda model: differential_sweep_specs
    ps = DifferentialParameterSweep(build_differential_sweep_specs=build_spec)
    ps.model_manager = ModelManager(ps)
    ps.model_manager.model = m
    local_values = np.array([0.0, 0.1, 1.0, 2.0])
    ps.diff_spec_index = [0, 1]
    diff_sweep_param_dict = ps._create_differential_sweep_params(local_values)

    expected_dict = {
        "fs.b": UniformSample(m.fs.input["b"], 0.11, 0.2, 1),
    }

    for key, value in diff_sweep_param_dict.items():
        assert value.lower_limit == expected_dict[key].lower_limit
        assert value.upper_limit == expected_dict[key].upper_limit


@pytest.mark.component
def test_bad_differential_sweep_specs(model, tmp_path):
    m = model
    differential_sweep_specs = {
        "fs.b": {
            "diff_mode": "percentile",
            "diff_sample_type": UniformSample,
            "relative_lb": 0.01,
            "relative_ub": 0.1,
            "nominal_lb": 0.0,
            "nominal_ub": 1.0,
            "pyomo_object": m.fs.input["b"],
        },
    }
    build_spec = lambda model: differential_sweep_specs
    ps = DifferentialParameterSweep(build_differential_sweep_specs=build_spec)
    ps.model_manager = ModelManager(ps)
    ps.model_manager.model = m
    with pytest.raises(ValueError):
        ps.parameter_sweep(
            build_model,
            build_sweep_params_for_tps,
            build_outputs=build_none_outputs,
            seed=0,
        )


@pytest.mark.component
def test_differential_sweep_outputs(model):
    m = model
    m.fs.slack_penalty = 1000.0
    m.fs.slack.setub(0)

    A = m.fs.input["a"]
    B = m.fs.input["b"]
    sweep_params = {A.name: (A, 0.1, 0.9, 3), B.name: (B, 0.0, 0.5, 3)}

    differential_sweep_specs = {
        B.name: {
            "diff_mode": "product",
            "diff_sample_type": UniformSample,
            "relative_lb": 0.01,
            "relative_ub": 0.01,
            "pyomo_object": m.fs.input["b"],
        },
    }

    outputs = {"fs.output[c]": m.fs.output["c"]}
    build_spec = lambda model: differential_sweep_specs
    ps = DifferentialParameterSweep(
        optimize_function=_optimization,
        reinitialize_function=_reinitialize,
        reinitialize_kwargs={"slack_penalty": 10.0},
        build_differential_sweep_specs=build_spec,
    )
    ps.model_manager = ModelManager(ps)
    ps.model_manager.model = m
    sweep_params, _ = ps._process_sweep_params(sweep_params)
    ps.config.differential_sweep_specs = build_spec(m)
    ps.config.build_outputs = lambda model: outputs
    ps._define_differential_sweep_outputs(model, sweep_params)

    # Finally test for the keys
    expected_keys = ["fs.output[c]", "fs.input[a]"]
    assert expected_keys == list(ps.differential_outputs.keys())


@pytest.mark.component
def test_differential_parameter_sweep(model, tmp_path):
    comm = MPI.COMM_WORLD
    tmp_path = _get_rank0_path(comm, tmp_path)

    results_fname = os.path.join(tmp_path, "global_results")
    csv_results_file_name = str(results_fname) + ".csv"
    h5_results_file_name = str(results_fname) + ".h5"

    m = model
    m.fs.slack_penalty = 10.0
    m.fs.slack.setub(None)

    A = m.fs.input["a"]
    B = m.fs.input["b"]
    differential_sweep_specs = {
        A.name: {
            "diff_mode": "sum",
            "diff_sample_type": NormalSample,
            "std_dev": 0.01,
            "pyomo_object": m.fs.input["a"],
        },
        B.name: {
            "diff_mode": "product",
            "diff_sample_type": UniformSample,
            "relative_lb": 0.01,
            "relative_ub": 0.01,
            "pyomo_object": m.fs.input["b"],
        },
    }

    ps = DifferentialParameterSweep(
        csv_results_file_name=csv_results_file_name,
        h5_results_file_name=h5_results_file_name,
        debugging_data_dir=None,  # Does not work at the moment
        interpolate_nan_outputs=True,
        optimize_function=_optimization,
        build_differential_sweep_specs=lambda model: differential_sweep_specs,
        initialize_function=_reinitialize,
        number_of_subprocesses=1,
    )

    # Call the parameter_sweep function
    _, global_results_dict = ps.parameter_sweep(
        build_model,
        build_sweep_params_for_tps,
        build_outputs=None,
        seed=0,
    )

    if ps.parallel_manager.is_root_process():
        truth_dict = {
            "differential_idx": np.array(
                [
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    0.0,
                    1.0,
                    2.0,
                    3.0,
                    4.0,
                    5.0,
                    6.0,
                    7.0,
                    8.0,
                ]
            ),
            "nominal_idx": np.array(
                [
                    0.0,
                    1.0,
                    2.0,
                    3.0,
                    4.0,
                    5.0,
                    6.0,
                    7.0,
                    8.0,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                ]
            ),
            "outputs": {
                "fs.input[a]": {
                    "full_name": "fs.input[a]",
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array(
                        [
                            0.1,
                            0.1,
                            0.1,
                            0.5,
                            0.5,
                            0.5,
                            0.9,
                            0.9,
                            0.9,
                            0.11764052,
                            0.11764052,
                            0.11764052,
                            0.51764052,
                            0.51764052,
                            0.51764052,
                            0.91764052,
                            0.91764052,
                            0.91764052,
                        ]
                    ),
                },
                "fs.input[b]": {
                    "full_name": "fs.input[b]",
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array(
                        [
                            0.0,
                            0.25,
                            0.5,
                            0.0,
                            0.25,
                            0.5,
                            0.0,
                            0.25,
                            0.5,
                            0.0,
                            0.0025,
                            0.005,
                            0.0,
                            0.0025,
                            0.005,
                            0.0,
                            0.0025,
                            0.005,
                        ]
                    ),
                },
                "fs.output[c]": {
                    "full_name": "fs.output[c]",
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array(
                        [
                            0.2,
                            0.2,
                            0.20000001,
                            1.00000001,
                            1.00000001,
                            1.00000001,
                            1.00000001,
                            1.00000001,
                            1.00000001,
                            0.23528105,
                            0.23528105,
                            0.23528106,
                            1.00000001,
                            1.00000001,
                            1.00000001,
                            1.00000001,
                            1.00000001,
                            1.00000001,
                        ]
                    ),
                },
                "fs.output[d]": {
                    "full_name": "fs.output[d]",
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array(
                        [
                            0.00000000e00,
                            7.50000000e-01,
                            1.00000001e00,
                            9.98996974e-09,
                            7.50000010e-01,
                            1.00000001e00,
                            9.92236517e-09,
                            7.50000010e-01,
                            1.00000001e00,
                            1.09912079e-22,
                            7.50000000e-03,
                            1.50000098e-02,
                            9.77884773e-09,
                            7.50000977e-03,
                            1.50000098e-02,
                            9.77884942e-09,
                            7.50000977e-03,
                            1.50000098e-02,
                        ]
                    ),
                },
                "fs.performance": {
                    "full_name": "fs.performance",
                    "value": np.array(
                        [
                            0.2,
                            0.95,
                            1.2,
                            1.0,
                            1.75,
                            2.0,
                            1.0,
                            1.75,
                            2.0,
                            0.23528105,
                            0.24278105,
                            0.25028107,
                            1.0,
                            1.0075,
                            1.0150,
                            1.0,
                            1.0075,
                            1.015,
                        ]
                    ),
                },
                "fs.slack[ab_slack]": {
                    "full_name": "fs.slack[ab_slack]",
                    "lower bound": 0,
                    "units": "None",
                    # "upper bound": 0,
                    "value": np.array(
                        [
                            0.0,
                            0.0,
                            -9.77208577e-09,
                            -9.54438112e-09,
                            -9.54432508e-09,
                            -9.54432509e-09,
                            0.8,
                            0.8,
                            0.8,
                            0.0,
                            0.0,
                            -9.77210838e-09,
                            3.52810371e-02,
                            3.52810371e-02,
                            3.52810371e-02,
                            8.35281037e-01,
                            8.35281037e-01,
                            8.35281037e-01,
                        ]
                    ),
                },
                "fs.slack[cd_slack]": {
                    "full_name": "fs.slack[cd_slack]",
                    "lower bound": 0,
                    "units": "None",
                    # "upper bound": 0,
                    "value": np.array(
                        [
                            0.0,
                            0.0,
                            0.5,
                            -9.98996974e-09,
                            -9.77226571e-09,
                            0.5,
                            -9.92236517e-09,
                            -9.77226571e-09,
                            0.5,
                            0.0,
                            0.0,
                            -9.77036119e-09,
                            -9.77884773e-09,
                            -9.76848697e-09,
                            -9.77036119e-09,
                            -9.77884942e-09,
                            -9.76848697e-09,
                            -9.77036119e-09,
                        ]
                    ),
                },
                "fs.slack_penalty": {
                    "full_name": "fs.slack_penalty",
                    "units": "None",
                    "value": np.array(
                        [
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                        ]
                    ),
                },
                "objective": {
                    "full_name": "objective",
                    "value": np.array(
                        [
                            0.2,
                            0.95,
                            -3.79999979,
                            1.00000021,
                            1.75000021,
                            -2.99999979,
                            -6.99999978,
                            -6.24999979,
                            -10.99999979,
                            0.23528105,
                            0.24278105,
                            0.25028126,
                            0.64718975,
                            0.65468975,
                            0.66218975,
                            -7.35281025,
                            -7.34531025,
                            -7.33781025,
                        ]
                    ),
                },
            },
            "solve_successful": [
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
            ],
            "sweep_params": {
                "fs.input[a]": {
                    "full_name": "fs.input[a]",
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array(
                        [
                            0.1,
                            0.1,
                            0.1,
                            0.5,
                            0.5,
                            0.5,
                            0.9,
                            0.9,
                            0.9,
                            0.11764052,
                            0.11764052,
                            0.11764052,
                            0.51764052,
                            0.51764052,
                            0.51764052,
                            0.91764052,
                            0.91764052,
                            0.91764052,
                        ]
                    ),
                },
                "fs.input[b]": {
                    "full_name": "fs.input[b]",
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array(
                        [
                            0.0,
                            0.25,
                            0.5,
                            0.0,
                            0.25,
                            0.5,
                            0.0,
                            0.25,
                            0.5,
                            0.0,
                            0.0025,
                            0.005,
                            0.0,
                            0.0025,
                            0.005,
                            0.0,
                            0.0025,
                            0.005,
                        ]
                    ),
                },
            },
        }

        read_dict = _read_output_h5(h5_results_file_name)
        _assert_h5_csv_agreement(csv_results_file_name, read_dict)
        # _assert_dictionary_correctness(global_results_dict, read_dict)
        if ps.parallel_manager.number_of_worker_processes() > 1:
            # Compare the sorted dictionary. We need to work with a sorted dictionary
            # because the differential parameter sweep produces a global dictionary
            # that is jumbled by the number of procs.
            if ps.parallel_manager.rank == 0:
                sorted_truth_dict = sort_output_dict(truth_dict)
                sorted_read_dict = sort_output_dict(read_dict)
                sorted_global_results_dict = sort_output_dict(global_results_dict)
                _assert_dictionary_correctness(sorted_truth_dict, sorted_read_dict)
                _assert_dictionary_correctness(
                    sorted_truth_dict, sorted_global_results_dict
                )
        else:
            _assert_dictionary_correctness(truth_dict, global_results_dict)
            _assert_dictionary_correctness(truth_dict, read_dict)


@pytest.mark.component
def test_differential_parameter_sweep_selective(model, tmp_path):
    comm = MPI.COMM_WORLD
    tmp_path = _get_rank0_path(comm, tmp_path)

    results_fname = os.path.join(tmp_path, "global_results")
    csv_results_file_name = str(results_fname) + ".csv"
    h5_results_file_name = str(results_fname) + ".h5"

    m = model
    m.fs.slack_penalty = 10.0
    m.fs.slack.setub(None)

    A = m.fs.input["a"]
    B = m.fs.input["b"]
    sweep_params = {A.name: (A, 0.1, 0.9, 3), B.name: (B, 0.0, 0.5, 3)}

    differential_sweep_specs = {
        B.name: {
            "diff_mode": "product",
            "diff_sample_type": UniformSample,
            "relative_lb": 0.25,
            "relative_ub": 0.75,
            "pyomo_object": m.fs.input["b"],
        },
    }
    build_spec = lambda model: differential_sweep_specs
    ps = DifferentialParameterSweep(
        csv_results_file_name=csv_results_file_name,
        h5_results_file_name=h5_results_file_name,
        debugging_data_dir=None,  # Does not work at the moment
        interpolate_nan_outputs=True,
        optimize_function=_optimization,
        build_differential_sweep_specs=build_spec,
        num_diff_samples=2,
    )

    # Call the parameter_sweep function
    outputs = {
        "fs.input[a]": m.fs.input["a"],
        "fs.input[b]": m.fs.input["b"],
        "fs.output[c]": m.fs.output["c"],
        "fs.output[d]": m.fs.output["d"],
        "fs.performance": m.fs.performance,
        "objective": m.objective,
        "fs.slack[ab_slack]": m.fs.slack["ab_slack"],
        "fs.slack[cd_slack]": m.fs.slack["cd_slack"],
        "fs.slack_penalty": m.fs.slack_penalty,
    }
    build_outputs = lambda model: outputs
    _, global_results_dict = ps.parameter_sweep(
        m,
        sweep_params,
        build_outputs=build_outputs,
        seed=0,
    )

    if ps.parallel_manager.is_root_process():
        truth_dict = {
            "differential_idx": np.array(
                [
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    0.0,
                    0.0,
                    1.0,
                    1.0,
                    2.0,
                    2.0,
                    3.0,
                    3.0,
                    4.0,
                    4.0,
                    5.0,
                    5.0,
                    6.0,
                    6.0,
                    7.0,
                    7.0,
                    8.0,
                    8.0,
                ]
            ),
            "nominal_idx": np.array(
                [
                    0.0,
                    1.0,
                    2.0,
                    3.0,
                    4.0,
                    5.0,
                    6.0,
                    7.0,
                    8.0,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                ]
            ),
            "outputs": {
                "fs.input[a]": {
                    "full_name": "fs.input[a]",
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array(
                        [
                            0.1,
                            0.1,
                            0.1,
                            0.5,
                            0.5,
                            0.5,
                            0.9,
                            0.9,
                            0.9,
                            0.1,
                            0.1,
                            0.1,
                            0.1,
                            0.1,
                            0.1,
                            0.5,
                            0.5,
                            0.5,
                            0.5,
                            0.5,
                            0.5,
                            0.9,
                            0.9,
                            0.9,
                            0.9,
                            0.9,
                            0.9,
                        ]
                    ),
                },
                "fs.input[b]": {
                    "full_name": "fs.input[b]",
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array(
                        [
                            0.0,
                            0.25,
                            0.5,
                            0.0,
                            0.25,
                            0.5,
                            0.0,
                            0.25,
                            0.5,
                            0.0,
                            0.0,
                            0.13110169,
                            0.15189867,
                            0.26220338,
                            0.30379734,
                            0.0,
                            0.0,
                            0.13110169,
                            0.15189867,
                            0.26220338,
                            0.30379734,
                            0.0,
                            0.0,
                            0.13110169,
                            0.15189867,
                            0.26220338,
                            0.30379734,
                        ]
                    ),
                },
                "fs.output[c]": {
                    "full_name": "fs.output[c]",
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array(
                        [
                            0.2,
                            0.2,
                            0.2,
                            1.0,
                            1.0,
                            1.0,
                            1.0,
                            1.0,
                            1.0,
                            0.2,
                            0.2,
                            0.2,
                            0.2,
                            0.2,
                            0.2,
                            1.0,
                            1.0,
                            1.0,
                            1.0,
                            1.0,
                            1.0,
                            1.0,
                            1.0,
                            1.0,
                            1.0,
                            1.0,
                            1.0,
                        ]
                    ),
                },
                "fs.output[d]": {
                    "full_name": "fs.output[d]",
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array(
                        [
                            0.0,
                            0.75,
                            1.0,
                            9.89148858e-09,
                            7.50000010e-01,
                            1.00000001e00,
                            9.77798881e-09,
                            7.50000010e-01,
                            1.0,
                            0.0,
                            0.0,
                            3.93305064e-01,
                            4.55696012e-01,
                            7.86610138e-01,
                            9.11392035e-01,
                            9.88979157e-09,
                            9.88979157e-09,
                            3.93305074e-01,
                            4.55696022e-01,
                            7.86610138e-01,
                            9.11392035e-01,
                            9.88962016e-09,
                            9.88962016e-09,
                            3.93305074e-01,
                            4.55696022e-01,
                            7.86610138e-01,
                            9.11392035e-01,
                        ]
                    ),
                },
                "fs.performance": {
                    "full_name": "fs.performance",
                    "value": np.array(
                        [
                            0.2,
                            0.95,
                            1.20,
                            1.00,
                            1.75,
                            2.0,
                            1.00,
                            1.75,
                            2.0,
                            0.2,
                            0.2,
                            0.59330506,
                            0.65569601,
                            0.98661015,
                            1.11139204,
                            1.00000002,
                            1.00000002,
                            1.39330508,
                            1.45569603,
                            1.78661015,
                            1.91139204,
                            1.0,
                            1.0,
                            1.39330508,
                            1.45569603,
                            1.78661015,
                            1.91139204,
                        ]
                    ),
                },
                "fs.slack[ab_slack]": {
                    "full_name": "fs.slack[ab_slack]",
                    "lower bound": 0,
                    "units": "None",
                    # "upper bound": 0,
                    "value": np.array(
                        [
                            0.0,
                            0.0,
                            -9.77208577e-09,
                            -9.54438112e-09,
                            -9.54432513e-09,
                            -9.54432506e-09,
                            0.8,
                            0.8,
                            0.8,
                            0.0,
                            0.0,
                            0.0,
                            0.0,
                            -9.77208594e-09,
                            -9.77208576e-09,
                            -9.54438113e-09,
                            -9.54438113e-09,
                            -9.54432505e-09,
                            -9.54432507e-09,
                            -9.54432502e-09,
                            -9.54432513e-09,
                            0.8,
                            0.8,
                            0.8,
                            0.8,
                            0.8,
                            0.8,
                        ]
                    ),
                },
                "fs.slack[cd_slack]": {
                    "full_name": "fs.slack[cd_slack]",
                    "lower bound": 0,
                    "units": "None",
                    # "upper bound": 0,
                    "value": np.array(
                        [
                            0.0,
                            0.0,
                            0.5,
                            -9.89148858e-09,
                            -9.77226571e-09,
                            0.5,
                            -9.77798881e-09,
                            -9.77226571e-09,
                            0.5,
                            0.0,
                            0.0,
                            0.0,
                            0.0,
                            -9.77228650e-09,
                            -9.77247711e-09,
                            -9.88979157e-09,
                            -9.88979157e-09,
                            -9.77216575e-09,
                            -9.77218082e-09,
                            -9.77228649e-09,
                            -9.77247712e-09,
                            -9.88962016e-09,
                            -9.88962016e-09,
                            -9.77216575e-09,
                            -9.77218081e-09,
                            -9.77228649e-09,
                            -9.77247712e-09,
                        ]
                    ),
                },
                "fs.slack_penalty": {
                    "full_name": "fs.slack_penalty",
                    "units": "None",
                    "value": np.array(
                        [
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                        ]
                    ),
                },
                "objective": {
                    "full_name": "objective",
                    "value": np.array(
                        [
                            0.2,
                            0.95,
                            -3.79999979,
                            1.0,
                            1.75,
                            -3.0,
                            -7.0,
                            -6.24999979,
                            -10.99999979,
                            0.2,
                            0.2,
                            0.59330506,
                            0.65569601,
                            0.98661034,
                            1.11139224,
                            1.0,
                            1.0,
                            1.39330528,
                            1.45569622,
                            1.78661034,
                            1.91139224,
                            -6.99999978,
                            -6.99999978,
                            -6.60669472,
                            -6.54430377,
                            -6.21338966,
                            -6.08860776,
                        ]
                    ),
                },
            },
            "solve_successful": [
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
            ],
            "sweep_params": {
                "fs.input[a]": {
                    "full_name": "fs.input[a]",
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array(
                        [
                            0.1,
                            0.1,
                            0.1,
                            0.5,
                            0.5,
                            0.5,
                            0.9,
                            0.9,
                            0.9,
                            0.1,
                            0.1,
                            0.1,
                            0.1,
                            0.1,
                            0.1,
                            0.5,
                            0.5,
                            0.5,
                            0.5,
                            0.5,
                            0.5,
                            0.9,
                            0.9,
                            0.9,
                            0.9,
                            0.9,
                            0.9,
                        ]
                    ),
                },
                "fs.input[b]": {
                    "full_name": "fs.input[b]",
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array(
                        [
                            0.0,
                            0.25,
                            0.5,
                            0.0,
                            0.25,
                            0.5,
                            0.0,
                            0.25,
                            0.5,
                            0.0,
                            0.0,
                            0.13110169,
                            0.15189867,
                            0.26220338,
                            0.30379734,
                            0.0,
                            0.0,
                            0.13110169,
                            0.15189867,
                            0.26220338,
                            0.30379734,
                            0.0,
                            0.0,
                            0.13110169,
                            0.15189867,
                            0.26220338,
                            0.30379734,
                        ]
                    ),
                },
            },
        }

        read_dict = _read_output_h5(h5_results_file_name)
        _assert_h5_csv_agreement(csv_results_file_name, read_dict)
        # _assert_dictionary_correctness(global_results_dict, read_dict)
        if ps.parallel_manager.number_of_worker_processes() > 1:
            # Compare the sorted dictionary. We need to work with a sorted dictionary
            # because the differential parameter sweep produces a global dictionary
            # that is jumbled by the number of procs.
            if ps.parallel_manager.rank == 0:
                sorted_truth_dict = sort_output_dict(truth_dict)
                sorted_read_dict = sort_output_dict(read_dict)
                sorted_global_results_dict = sort_output_dict(global_results_dict)
                _assert_dictionary_correctness(sorted_truth_dict, sorted_read_dict)
                _assert_dictionary_correctness(
                    sorted_truth_dict, sorted_global_results_dict
                )
        else:
            _assert_dictionary_correctness(truth_dict, global_results_dict)
            _assert_dictionary_correctness(truth_dict, read_dict)


@pytest.mark.component
def test_differential_parameter_sweep_function(model, tmp_path):
    comm = MPI.COMM_WORLD
    tmp_path = _get_rank0_path(comm, tmp_path)

    results_fname = os.path.join(tmp_path, "global_results")
    csv_results_file_name = str(results_fname) + ".csv"
    h5_results_file_name = str(results_fname) + ".h5"

    m = model
    m.fs.slack_penalty = 10.0
    m.fs.slack.setub(None)

    A = m.fs.input["a"]
    B = m.fs.input["b"]
    sweep_params = {A.name: (A, 0.1, 0.9, 3), B.name: (B, 0.0, 0.5, 3)}

    differential_sweep_specs = {
        A.name: {
            "diff_mode": "sum",
            "diff_sample_type": NormalSample,
            "std_dev": 0.01,
            "pyomo_object": m.fs.input["a"],
        },
        B.name: {
            "diff_mode": "product",
            "diff_sample_type": UniformSample,
            "relative_lb": 0.01,
            "relative_ub": 0.01,
            "pyomo_object": m.fs.input["b"],
        },
    }

    # Call the parameter_sweep function
    _, global_results_dict = differential_parameter_sweep(
        build_model,
        build_sweep_params_for_tps,
        differential_sweep_specs,
        build_outputs=None,
        csv_results_file_name=csv_results_file_name,
        h5_results_file_name=h5_results_file_name,
        debugging_data_dir=None,  # Does not work at the moment
        interpolate_nan_outputs=True,
        optimize_function=_optimization,
        initialize_function=_reinitialize,
        seed=0,
    )

    if comm.rank == 0:
        truth_dict = {
            "differential_idx": np.array(
                [
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    0.0,
                    1.0,
                    2.0,
                    3.0,
                    4.0,
                    5.0,
                    6.0,
                    7.0,
                    8.0,
                ]
            ),
            "nominal_idx": np.array(
                [
                    0.0,
                    1.0,
                    2.0,
                    3.0,
                    4.0,
                    5.0,
                    6.0,
                    7.0,
                    8.0,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                ]
            ),
            "outputs": {
                "fs.input[a]": {
                    "full_name": "fs.input[a]",
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array(
                        [
                            0.1,
                            0.1,
                            0.1,
                            0.5,
                            0.5,
                            0.5,
                            0.9,
                            0.9,
                            0.9,
                            0.11764052,
                            0.11764052,
                            0.11764052,
                            0.51764052,
                            0.51764052,
                            0.51764052,
                            0.91764052,
                            0.91764052,
                            0.91764052,
                        ]
                    ),
                },
                "fs.input[b]": {
                    "full_name": "fs.input[b]",
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array(
                        [
                            0.0,
                            0.25,
                            0.5,
                            0.0,
                            0.25,
                            0.5,
                            0.0,
                            0.25,
                            0.5,
                            0.0,
                            0.0025,
                            0.005,
                            0.0,
                            0.0025,
                            0.005,
                            0.0,
                            0.0025,
                            0.005,
                        ]
                    ),
                },
                "fs.output[c]": {
                    "full_name": "fs.output[c]",
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array(
                        [
                            0.2,
                            0.2,
                            0.2,
                            1.0,
                            1.0,
                            1.0,
                            1.0,
                            1.0,
                            1.0,
                            0.23528105,
                            0.23528105,
                            0.23528106,
                            1.0,
                            1.0,
                            1.0,
                            1.0,
                            1.0,
                            1.0,
                        ]
                    ),
                },
                "fs.output[d]": {
                    "full_name": "fs.output[d]",
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array(
                        [
                            0.0,
                            0.75,
                            1.0,
                            9.98996974e-09,
                            0.75,
                            1.0,
                            9.92236517e-09,
                            7.50000010e-01,
                            1.0,
                            1.09912079e-22,
                            7.50000000e-03,
                            1.50000098e-02,
                            9.77884773e-09,
                            7.50000977e-03,
                            1.50000098e-02,
                            9.77884942e-09,
                            7.50000977e-03,
                            1.50000098e-02,
                        ]
                    ),
                },
                "fs.performance": {
                    "full_name": "fs.performance",
                    "value": np.array(
                        [
                            0.2,
                            0.95,
                            1.2,
                            1.0,
                            1.75,
                            2.0,
                            1.0,
                            1.75,
                            2.0,
                            0.23528105,
                            0.24278105,
                            0.25028107,
                            1.0,
                            1.0075,
                            1.0150,
                            1.0,
                            1.0075,
                            1.015,
                        ]
                    ),
                },
                "fs.slack[ab_slack]": {
                    "full_name": "fs.slack[ab_slack]",
                    "lower bound": 0,
                    "units": "None",
                    # "upper bound": 0,
                    "value": np.array(
                        [
                            0.0,
                            0.0,
                            -9.77208577e-09,
                            -9.54438112e-09,
                            -9.54432508e-09,
                            -9.54432509e-09,
                            0.8,
                            0.8,
                            0.8,
                            0.0,
                            0.0,
                            -9.77210838e-09,
                            3.52810371e-02,
                            3.52810371e-02,
                            3.52810371e-02,
                            8.35281037e-01,
                            8.35281037e-01,
                            8.35281037e-01,
                        ]
                    ),
                },
                "fs.slack[cd_slack]": {
                    "full_name": "fs.slack[cd_slack]",
                    "lower bound": 0,
                    "units": "None",
                    # "upper bound": 0,
                    "value": np.array(
                        [
                            0.0,
                            0.0,
                            0.5,
                            -9.98996974e-09,
                            -9.77226571e-09,
                            4.99999990e-01,
                            -9.92236517e-09,
                            -9.77226571e-09,
                            4.99999990e-01,
                            0.0,
                            0.0,
                            -9.77036119e-09,
                            -9.77884773e-09,
                            -9.76848697e-09,
                            -9.77036119e-09,
                            -9.77884942e-09,
                            -9.76848697e-09,
                            -9.77036119e-09,
                        ]
                    ),
                },
                "fs.slack_penalty": {
                    "full_name": "fs.slack_penalty",
                    "units": "None",
                    "value": np.array(
                        [
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                        ]
                    ),
                },
                "objective": {
                    "full_name": "objective",
                    "value": np.array(
                        [
                            0.2,
                            0.95,
                            -3.8,
                            1.0,
                            1.75,
                            -3.0,
                            -6.99999978,
                            -6.24999979,
                            -10.99999979,
                            0.23528105,
                            0.24278105,
                            0.25028126,
                            0.64718975,
                            0.65468975,
                            0.66218975,
                            -7.35281025,
                            -7.34531025,
                            -7.33781025,
                        ]
                    ),
                },
            },
            "solve_successful": [
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
            ],
            "sweep_params": {
                "fs.input[a]": {
                    "full_name": "fs.input[a]",
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array(
                        [
                            0.1,
                            0.1,
                            0.1,
                            0.5,
                            0.5,
                            0.5,
                            0.9,
                            0.9,
                            0.9,
                            0.11764052,
                            0.11764052,
                            0.11764052,
                            0.51764052,
                            0.51764052,
                            0.51764052,
                            0.91764052,
                            0.91764052,
                            0.91764052,
                        ]
                    ),
                },
                "fs.input[b]": {
                    "full_name": "fs.input[b]",
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array(
                        [
                            0.0,
                            0.25,
                            0.5,
                            0.0,
                            0.25,
                            0.5,
                            0.0,
                            0.25,
                            0.5,
                            0.0,
                            0.0025,
                            0.005,
                            0.0,
                            0.0025,
                            0.005,
                            0.0,
                            0.0025,
                            0.005,
                        ]
                    ),
                },
            },
        }

        read_dict = _read_output_h5(h5_results_file_name)
        _assert_h5_csv_agreement(csv_results_file_name, read_dict)
        if comm.size > 1:
            # Compare the sorted dictionary. We need to work with a sorted dictionary
            # because the differential parameter sweep produces a global dictionary
            # that is jumbled by the number of procs.
            if comm.rank == 0:
                sorted_truth_dict = sort_output_dict(truth_dict)
                sorted_read_dict = sort_output_dict(read_dict)
                sorted_global_results_dict = sort_output_dict(global_results_dict)
                _assert_dictionary_correctness(sorted_truth_dict, sorted_read_dict)
                _assert_dictionary_correctness(
                    sorted_truth_dict, sorted_global_results_dict
                )
        else:
            _assert_dictionary_correctness(truth_dict, global_results_dict)
            _assert_dictionary_correctness(truth_dict, read_dict)


def sort_output_dict(input_dict):
    """Simple utility function to sort all values in ascending order"""

    sorted_dict = copy.deepcopy(input_dict)
    for key, item in input_dict.items():
        if key in ["sweep_params", "outputs"]:  # != "solve_successful":
            for subkey, subitem in item.items():
                sorted_dict[key][subkey]["value"] = np.sort(subitem["value"])
        elif key == "solve_successful":
            sorted_dict["solve_successful"] = np.sort(input_dict[key]).tolist()
        elif key in ["nominal_idx", "differential_idx"]:
            sorted_dict[key] = np.sort(item)

    return sorted_dict



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
#################################################################################

import pytest

from watertap.tools.oli_api.util.fixed_keys_dict import (
    default_oli_input_dict,
)


@pytest.mark.unit
def test_fixed_keys_dict():

    with pytest.raises(RuntimeError):
        default_oli_input_dict["invalid_key"] = "value"

    with pytest.raises(Exception):
        del default_oli_input_dict["any_key"]

    with pytest.raises(RuntimeError):
        default_oli_input_dict._check_value("temperature_unit", ["not_K"])

    default_oli_input_dict.pprint()



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import os
import pytest
import tempfile
from watertap.examples.flowsheets.case_studies.wastewater_resource_recovery.amo_1690.amo_1690_sweep import (
    run_analysis,
)

sweep_list = [
    1,
    2,
    3,
]


@pytest.mark.parametrize("case_num", sweep_list)
@pytest.mark.integration
def test_sweep(case_num, tmp_path):
    cwd = os.getcwd()
    os.chdir(tmp_path)
    temp = tempfile.NamedTemporaryFile(delete=False)
    temp.close()
    global_results, sweep_params, m = run_analysis(
        case_num,
        nx=2,
        interpolate_nan_outputs=False,
        save_path=temp.name,
    )
    os.remove(temp.name)
    os.chdir(cwd)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pytest
import watertap.property_models.NDMA_prop_pack as props
from idaes.models.properties.tests.test_harness import (
    PropertyTestHarness as PropertyTestHarness_idaes,
)
from watertap.property_models.tests.property_test_harness import (
    PropertyTestHarness,
    PropertyRegressionTest,
    PropertyCalculateStateTest,
)

# -----------------------------------------------------------------------------
@pytest.mark.unit
class TestNDMAroperty_idaes(PropertyTestHarness_idaes):
    def configure(self):
        self.prop_pack = props.NDMAParameterBlock
        self.param_args = {}
        self.prop_args = {}
        self.has_density_terms = False


class TestNDMAProperty(PropertyTestHarness):
    def configure(self):
        self.prop_pack = props.NDMAParameterBlock
        self.param_args = {}
        self.scaling_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 1,
            ("flow_mass_phase_comp", ("Liq", "NDMA")): 1e8,
        }
        self.stateblock_statistics = {
            "number_variables": 15,
            "number_total_constraints": 11,
            "number_unused_variables": 2,  # pressure is unused
            "default_degrees_of_freedom": 2,
        }  # 3 state vars, but pressure is not active
        self.default_solution = {
            ("mass_frac_phase_comp", ("Liq", "H2O")): 0.999999926,
            ("mass_frac_phase_comp", ("Liq", "NDMA")): 74e-9,
            ("dens_mass_phase", "Liq"): 997,
            ("flow_vol_phase", "Liq"): 1.003e-3,
            ("conc_mass_phase_comp", ("Liq", "H2O")): 997,
            ("conc_mass_phase_comp", ("Liq", "NDMA")): 7.3778e-5,
            ("flow_mol_phase_comp", ("Liq", "H2O")): 55.5084,
            ("flow_mol_phase_comp", ("Liq", "NDMA")): 9.9889e-7,
            ("mole_frac_phase_comp", ("Liq", "H2O")): 0.999999982,
            ("mole_frac_phase_comp", ("Liq", "NDMA")): 1.7995e-8,
            ("molality_phase_comp", ("Liq", "NDMA")): 9.9889e-7,
        }


@pytest.mark.component
class TestNDMAPropertySolution_1(PropertyRegressionTest):
    def configure(self):
        self.prop_pack = props.NDMAParameterBlock
        self.param_args = {}

        self.solver = "ipopt"
        self.optarg = {"nlp_scaling_method": "user-scaling"}

        self.scaling_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 1,
            ("flow_mass_phase_comp", ("Liq", "NDMA")): 1e2,
        }
        self.state_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 0.95,
            ("flow_mass_phase_comp", ("Liq", "NDMA")): 0.05,
            ("temperature", None): 273.15 + 25,
            ("pressure", None): 1e5,
        }
        self.regression_solution = {
            ("mass_frac_phase_comp", ("Liq", "H2O")): 0.95,
            ("mass_frac_phase_comp", ("Liq", "NDMA")): 0.05,
            ("dens_mass_phase", "Liq"): 997,
            ("flow_vol_phase", "Liq"): 1.003e-3,
            ("conc_mass_phase_comp", ("Liq", "H2O")): 947.15,
            ("conc_mass_phase_comp", ("Liq", "NDMA")): 49.85,
            ("flow_mol_phase_comp", ("Liq", "H2O")): 52.73,
            ("flow_mol_phase_comp", ("Liq", "NDMA")): 0.6749,
            ("mole_frac_phase_comp", ("Liq", "H2O")): 0.9874,
            ("mole_frac_phase_comp", ("Liq", "NDMA")): 1.2637e-2,
            ("molality_phase_comp", ("Liq", "NDMA")): 0.7105,
        }


@pytest.mark.component
class TestNDMACalculateState_1(PropertyCalculateStateTest):
    def configure(self):
        self.prop_pack = props.NDMAParameterBlock
        self.param_args = {}

        self.solver = "ipopt"
        self.optarg = {"nlp_scaling_method": "user-scaling"}

        self.scaling_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 1e-1,
            ("flow_mass_phase_comp", ("Liq", "NDMA")): 1e1,
        }
        self.var_args = {
            ("flow_vol_phase", "Liq"): 2e-2,
            ("mass_frac_phase_comp", ("Liq", "NDMA")): 0.05,
            ("temperature", None): 273.15 + 25,
            ("pressure", None): 1e5,
        }
        self.state_solution = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 18.943,
            ("flow_mass_phase_comp", ("Liq", "NDMA")): 0.997,
        }



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
Translator block representing the ASM1/ADM1 interface.

Assumptions:
     * Steady-state only

Model formulated from:

Copp J. and Jeppsson, U., Rosen, C., 2006.
Towards an ASM1 - ADM1 State Variable Interface for Plant-Wide Wastewater Treatment Modeling.
Proceedings of the Water Environment Federation, 2003, pp 498-510.
"""

# Import Pyomo libraries
from pyomo.common.config import ConfigBlock, ConfigValue

# Import IDAES cores
from idaes.core.util.math import smooth_max
from idaes.core import declare_process_block_class
from idaes.models.unit_models.translator import TranslatorData
from idaes.core.util.config import (
    is_reaction_parameter_block,
)
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.solvers import get_solver
import idaes.logger as idaeslog
import idaes.core.util.scaling as iscale

from pyomo.environ import (
    Param,
    NonNegativeReals,
    Var,
    units as pyunits,
    check_optimal_termination,
    Set,
    Expr_if,
)

from idaes.core.util.exceptions import InitializationError

__author__ = "Alejandro Garciadiego, Xinhong Liu"


# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("Translator_ASM1_ADM1")
class TranslatorDataASM1ADM1(TranslatorData):
    """
    Translator block representing the ASM1/ADM1 interface
    """

    CONFIG = TranslatorData.CONFIG()
    CONFIG.declare(
        "reaction_package",
        ConfigValue(
            default=None,
            domain=is_reaction_parameter_block,
            description="Reaction package to use for control volume",
            doc="""Reaction parameter object used to define reaction calculations,
**default** - None.
**Valid values:** {
**None** - no reaction package,
**ReactionParameterBlock** - a ReactionParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "reaction_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing reaction packages",
            doc="""A ConfigBlock with arguments to be passed to a reaction block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see reaction package for documentation.}""",
        ),
    )

    def build(self):
        """
        Begin building model.
        Args:
            None
        Returns:
            None
        """
        # Call UnitModel.build to setup dynamics
        super(TranslatorDataASM1ADM1, self).build()

        self.i_xe = Param(
            initialize=0.06,
            units=pyunits.dimensionless,
            mutable=True,
            doc="Fraction nitrogen in particulate products",
        )

        self.i_xb = Param(
            initialize=0.08,
            units=pyunits.dimensionless,
            mutable=True,
            doc="Fraction nitrogen in biomass",
        )

        self.f_xI = Param(
            initialize=0.05,
            units=pyunits.dimensionless,
            mutable=True,
            doc="Anaerobic degradable fraction of X_I and X_P",
        )

        self.eps = Param(
            initialize=1e-10,
            units=pyunits.kg / pyunits.m**3,
            mutable=True,
            doc="Smoothing factor",
        )

        mw_n = 14 * pyunits.kg / pyunits.kmol
        mw_c = 12 * pyunits.kg / pyunits.kmol

        @self.Constraint(
            self.flowsheet().time,
            doc="Equality volumetric flow equation",
        )
        def eq_flow_vol_rule(blk, t):
            return blk.properties_out[t].flow_vol == blk.properties_in[t].flow_vol

        @self.Constraint(
            self.flowsheet().time,
            doc="Equality temperature equation",
        )
        def eq_temperature_rule(blk, t):
            return blk.properties_out[t].temperature == blk.properties_in[t].temperature

        @self.Constraint(
            self.flowsheet().time,
            doc="Equality pressure equation",
        )
        def eq_pressure_rule(blk, t):
            return blk.properties_out[t].pressure == blk.properties_in[t].pressure

        @self.Expression(self.flowsheet().time, doc="COD demand")
        def CODd(blk, t):
            return (
                blk.properties_in[t].conc_mass_comp["S_O"]
                + blk.properties_in[t].conc_mass_comp["S_NO"] * 2.86
            )

        self.inter_S_S = Var(
            self.flowsheet().time,
            initialize=self.properties_in[0].conc_mass_comp["S_S"],
            units=pyunits.kg / pyunits.m**3,
            domain=NonNegativeReals,
            doc="Readily biodegradable substrate remaining",
        )

        @self.Constraint(
            self.flowsheet().time,
            doc="COD demand minus readily biodegradable substrate",
        )
        def CODd_step1(blk, t):
            return blk.inter_S_S[t] == smooth_max(
                0 * pyunits.kg / pyunits.m**3,
                blk.properties_in[t].conc_mass_comp["S_S"] - blk.CODd[t],
                blk.eps,
            )

        @self.Expression(self.flowsheet().time, doc="COD demand after S_S")
        # Includes smooth formulation for max(x, 0) to avoid having negative values
        # formulation included in all subsequent values for COD requirements and
        # COD remains
        def CODd2(blk, t):
            return smooth_max(
                0 * pyunits.kg / pyunits.m**3,
                blk.CODd[t] - blk.properties_in[t].conc_mass_comp["S_S"],
                blk.eps,
            )

        self.inter_X_S = Var(
            self.flowsheet().time,
            initialize=self.properties_in[0].conc_mass_comp["X_S"],
            units=pyunits.kg / pyunits.m**3,
            domain=NonNegativeReals,
            doc="Slowly biodegradable substrate remaining",
        )

        @self.Constraint(
            self.flowsheet().time,
            doc="COD demand minus slowly biodegradable substrate",
        )
        def CODd_step2(blk, t):
            return blk.inter_X_S[t] == smooth_max(
                0 * pyunits.kg / pyunits.m**3,
                blk.properties_in[t].conc_mass_comp["X_S"] - blk.CODd2[t],
                blk.eps,
            )

        @self.Expression(self.flowsheet().time, doc="COD demand after X_S")
        def CODd3(blk, t):
            return smooth_max(
                0 * pyunits.kg / pyunits.m**3,
                blk.CODd2[t] - blk.properties_in[t].conc_mass_comp["X_S"],
                blk.eps,
            )

        self.inter_X_BH = Var(
            self.flowsheet().time,
            initialize=self.properties_in[0].conc_mass_comp["X_BH"],
            units=pyunits.kg / pyunits.m**3,
            domain=NonNegativeReals,
            doc="Active heterotrophic biomass remaining",
        )

        @self.Constraint(
            self.flowsheet().time,
            doc="COD demand minus active heterotrophic biomass",
        )
        def CODd_step3(blk, t):
            return blk.inter_X_BH[t] == smooth_max(
                0 * pyunits.kg / pyunits.m**3,
                blk.properties_in[t].conc_mass_comp["X_BH"] - blk.CODd3[t],
                blk.eps,
            )

        @self.Expression(self.flowsheet().time, doc="COD demand after X_BH")
        def CODd4(blk, t):
            return smooth_max(
                0 * pyunits.kg / pyunits.m**3,
                blk.CODd3[t] - blk.properties_in[t].conc_mass_comp["X_BH"],
                blk.eps,
            )

        self.inter_X_BA = Var(
            self.flowsheet().time,
            initialize=self.properties_in[0].conc_mass_comp["X_BA"],
            units=pyunits.kg / pyunits.m**3,
            domain=NonNegativeReals,
            doc="Active autotrophic biomass remaining",
        )

        @self.Constraint(
            self.flowsheet().time,
            doc="COD demand minus active autotrophic biomass",
        )
        def CODd_step4(blk, t):
            return blk.inter_X_BA[t] == smooth_max(
                0 * pyunits.kg / pyunits.m**3,
                blk.properties_in[t].conc_mass_comp["X_BA"] - blk.CODd4[t],
                blk.eps,
            )

        @self.Expression(self.flowsheet().time, doc="COD demand after X_BA")
        def CODd5(blk, t):
            return smooth_max(
                0 * pyunits.kg / pyunits.m**3,
                blk.CODd4[t] - blk.properties_in[t].conc_mass_comp["X_BA"],
                blk.eps,
            )

        @self.Expression(self.flowsheet().time, doc="Soluble COD")
        def CODs(blk, t):
            return blk.properties_in[t].conc_mass_comp["S_I"] + blk.inter_S_S[t]

        @self.Expression(self.flowsheet().time, doc="Particulate COD")
        def CODp(blk, t):
            return (
                blk.properties_in[t].conc_mass_comp["X_I"]
                + blk.inter_X_S[t]
                + blk.inter_X_BH[t]
                + blk.inter_X_BA[t]
                + blk.properties_in[t].conc_mass_comp["X_P"]
            )

        @self.Expression(self.flowsheet().time, doc="Total COD")
        def CODt(blk, t):
            return blk.CODs[t] + blk.CODp[t]

        self.TKN_in = Var(
            self.flowsheet().time,
            initialize=self.properties_in[0].conc_mass_comp["S_NH"]
            + self.properties_in[0].conc_mass_comp["S_ND"]
            + self.properties_in[0].conc_mass_comp["X_ND"]
            + self.i_xb
            * (
                self.properties_in[0].conc_mass_comp["X_BH"]
                + self.properties_in[0].conc_mass_comp["X_BA"]
            )
            + self.i_xe
            * (
                self.properties_in[0].conc_mass_comp["X_I"]
                + self.properties_in[0].conc_mass_comp["X_P"]
            ),
            units=pyunits.kg / pyunits.m**3,
            domain=NonNegativeReals,
            doc="Total Kjeldahl nitrogen",
        )

        @self.Constraint(
            self.flowsheet().time,
            doc="Total Kjeldahl nitrogen",
        )
        def TKNin(blk, t):
            return blk.TKN_in[t] == blk.properties_in[t].conc_mass_comp[
                "S_NH"
            ] + blk.properties_in[t].conc_mass_comp["S_ND"] + blk.properties_in[
                t
            ].conc_mass_comp[
                "X_ND"
            ] + self.i_xb * (
                blk.inter_X_BH[t] + blk.inter_X_BA[t]
            ) + self.i_xe * (
                blk.properties_in[t].conc_mass_comp["X_I"]
                + blk.properties_in[t].conc_mass_comp["X_P"]
            )

        @self.Expression(self.flowsheet().time, doc="Required soluble COD")
        def ReqCODs(blk, t):
            return (
                blk.properties_in[t].conc_mass_comp["S_ND"]
                / blk.config.reaction_package.N_aa
                / mw_n
            )

        @self.Expression(self.flowsheet().time, doc="Amino acids mapping")
        def saa_mapping(blk, t):
            return Expr_if(
                blk.inter_S_S[t] > blk.ReqCODs[t],
                blk.ReqCODs[t],
                blk.inter_S_S[t],
            )

        @self.Expression(self.flowsheet().time, doc="Monosaccharides mapping step A")
        def ssu_mapping_A(blk, t):
            return Expr_if(
                blk.inter_S_S[t] > blk.ReqCODs[t],
                blk.inter_S_S[t] - blk.ReqCODs[t],
                1e-10 * pyunits.kg / pyunits.m**3,
            )

        @self.Constraint(
            self.flowsheet().time,
            doc="Amino acids balance",
        )
        def Saa_balance(blk, t):
            return blk.properties_out[t].conc_mass_comp["S_aa"] == blk.saa_mapping[t]

        @self.Expression(self.flowsheet().time, doc="COD remaining from step A")
        def COD_remain_a(blk, t):
            return smooth_max(
                0 * pyunits.kg / pyunits.m**3, blk.CODt[t] - blk.inter_S_S[t], blk.eps
            )

        @self.Expression(self.flowsheet().time, doc="Organic nitrogen pool from step A")
        def ORGN_remain_a(blk, t):
            return (
                blk.TKN_in[t]
                - (
                    blk.properties_out[t].conc_mass_comp["S_aa"]
                    * blk.config.reaction_package.N_aa
                    * mw_n
                )
                - blk.properties_in[t].conc_mass_comp["S_NH"]
            )

        @self.Expression(
            self.flowsheet().time, doc="Required soluble inert organic nitrogen"
        )
        def ReqOrgNS(blk, t):
            return (
                blk.properties_in[t].conc_mass_comp["S_I"]
                * blk.config.reaction_package.N_I
                * mw_n
            )

        @self.Expression(self.flowsheet().time, doc="Soluble inert mapping step B")
        def si_mapping_B(blk, t):
            return Expr_if(
                blk.ORGN_remain_a[t] > blk.ReqOrgNS[t],
                blk.properties_in[t].conc_mass_comp["S_I"],
                blk.ORGN_remain_a[t] / blk.config.reaction_package.N_I / mw_n,
            )

        @self.Expression(self.flowsheet().time, doc="Monosaccharides mapping step B")
        def ssu_mapping_B(blk, t):
            return Expr_if(
                blk.ORGN_remain_a[t] > blk.ReqOrgNS[t],
                blk.ssu_mapping_A[t],
                blk.ssu_mapping_A[t]
                + blk.properties_in[t].conc_mass_comp["S_I"]
                - blk.si_mapping_B[t],
            )

        @self.Constraint(
            self.flowsheet().time,
            doc="Inert organic balance",
        )
        def Si_balance(blk, t):
            return blk.properties_out[t].conc_mass_comp["S_I"] == blk.si_mapping_B[t]

        @self.Constraint(
            self.flowsheet().time,
            doc="Monosacharides balance",
        )
        def Ssu_balance(blk, t):
            return blk.properties_out[t].conc_mass_comp["S_su"] == blk.ssu_mapping_B[t]

        @self.Expression(self.flowsheet().time, doc="COD remaining from step B")
        def COD_remain_b(blk, t):
            return smooth_max(
                0 * pyunits.kg / pyunits.m**3,
                blk.COD_remain_a[t] - blk.properties_in[t].conc_mass_comp["S_I"],
                blk.eps,
            )

        @self.Expression(self.flowsheet().time, doc="Organic nitrogen pool from step B")
        def ORGN_remain_b(blk, t):
            return blk.ORGN_remain_a[t] - (
                blk.properties_out[t].conc_mass_comp["S_I"]
                * blk.config.reaction_package.N_I
                * mw_n
            )

        @self.Expression(
            self.flowsheet().time, doc="Required particulate inert material"
        )
        def ReqOrgNx(blk, t):
            return (
                blk.f_xI
                * (
                    blk.properties_in[t].conc_mass_comp["X_P"]
                    + blk.properties_in[t].conc_mass_comp["X_I"]
                )
                * blk.config.reaction_package.N_I
                * mw_n
            )

        @self.Expression(
            self.flowsheet().time,
            doc="Inert particulate organic material mapping step C",
        )
        def xi_mapping(blk, t):
            return Expr_if(
                blk.ORGN_remain_b[t] > blk.ReqOrgNx[t],
                (
                    blk.f_xI
                    * (
                        blk.properties_in[t].conc_mass_comp["X_P"]
                        + blk.properties_in[t].conc_mass_comp["X_I"]
                    )
                ),
                blk.ORGN_remain_b[t] / blk.config.reaction_package.N_I / mw_n,
            )

        @self.Constraint(
            self.flowsheet().time,
            doc="Organic nitrogen balance",
        )
        def Xi_balance(blk, t):
            return blk.properties_out[t].conc_mass_comp["X_I"] == blk.xi_mapping[t]

        @self.Expression(self.flowsheet().time, doc="COD remaining from step C")
        def COD_remain_c(blk, t):
            return smooth_max(
                0 * pyunits.kg / pyunits.m**3,
                blk.COD_remain_b[t] - blk.properties_out[t].conc_mass_comp["X_I"],
                blk.eps,
            )

        @self.Expression(self.flowsheet().time, doc="Organic nitrogen pool from step C")
        def ORGN_remain_c(blk, t):
            return blk.ORGN_remain_b[t] - (
                blk.properties_out[t].conc_mass_comp["X_I"]
                * blk.config.reaction_package.N_I
                * mw_n
            )

        @self.Expression(self.flowsheet().time, doc="Required Composites")
        def ReqCODXc(blk, t):
            return blk.ORGN_remain_c[t] / blk.config.reaction_package.N_xc / mw_n

        @self.Expression(self.flowsheet().time, doc="Composites mapping")
        def xc_mapping(blk, t):
            return Expr_if(
                blk.COD_remain_c[t] > blk.ReqCODXc[t],
                blk.ReqCODXc[t],
                blk.COD_remain_c[t],
            )

        @self.Constraint(
            self.flowsheet().time,
            doc="Composites balance",
        )
        def Xc_balance(blk, t):
            return blk.properties_out[t].conc_mass_comp["X_c"] == blk.xc_mapping[t]

        @self.Expression(self.flowsheet().time, doc="Carbohydrates mapping")
        def xch_mapping(blk, t):
            return Expr_if(
                blk.COD_remain_c[t] > blk.ReqCODXc[t],
                blk.config.reaction_package.f_ch_xc
                / (
                    blk.config.reaction_package.f_ch_xc
                    + blk.config.reaction_package.f_li_xc
                )
                * (blk.COD_remain_c[t] - blk.properties_out[t].conc_mass_comp["X_c"]),
                1e-10 * pyunits.kg / pyunits.m**3,
            )

        @self.Constraint(
            self.flowsheet().time,
            doc="Carbohydrates balance",
        )
        def xch_balance(blk, t):
            return blk.properties_out[t].conc_mass_comp["X_ch"] == blk.xch_mapping[t]

        @self.Expression(self.flowsheet().time, doc="Lipids mapping")
        def xli_mapping(blk, t):
            return Expr_if(
                blk.COD_remain_c[t] > blk.ReqCODXc[t],
                blk.config.reaction_package.f_li_xc
                / (
                    blk.config.reaction_package.f_ch_xc
                    + blk.config.reaction_package.f_li_xc
                )
                * (blk.COD_remain_c[t] - blk.properties_out[t].conc_mass_comp["X_c"]),
                1e-10 * pyunits.kg / pyunits.m**3,
            )

        @self.Constraint(
            self.flowsheet().time,
            doc="Lipids balance",
        )
        def xli_balance(blk, t):
            return blk.properties_out[t].conc_mass_comp["X_li"] == blk.xli_mapping[t]

        @self.Expression(self.flowsheet().time, doc="Inorganic nitrogen mapping")
        def sin_mapping(blk, t):
            return Expr_if(
                blk.COD_remain_c[t] > blk.ReqCODXc[t],
                blk.properties_in[t].conc_mass_comp["S_NH"],
                blk.properties_in[t].conc_mass_comp["S_NH"]
                + (
                    blk.ORGN_remain_c[t]
                    - blk.properties_out[t].conc_mass_comp["X_c"]
                    * blk.config.reaction_package.N_xc
                    * mw_n
                ),
            )

        @self.Constraint(
            self.flowsheet().time,
            doc="Inorganic nitrogen balance",
        )
        def SIN_balance(blk, t):
            return blk.properties_out[t].conc_mass_comp["S_IN"] == blk.sin_mapping[t]

        @self.Constraint(
            self.flowsheet().time,
            doc="Anions balance",
        )
        def return_an(blk, t):
            return (
                blk.properties_out[t].anions
                == blk.properties_out[t].conc_mass_comp["S_IN"] / mw_n
            )

        @self.Constraint(
            self.flowsheet().time,
            doc="Cations balance",
        )
        def return_cat(blk, t):
            return (
                blk.properties_out[t].cations
                == blk.properties_out[t].conc_mass_comp["S_IC"] / mw_c
            )

        self.zero_flow_components = Set(
            initialize=[
                "S_fa",
                "S_va",
                "S_bu",
                "S_pro",
                "S_ac",
                "S_h2",
                "S_ch4",
                "X_pr",
                "X_su",
                "X_aa",
                "X_fa",
                "X_c4",
                "X_pro",
                "X_ac",
                "X_h2",
            ]
        )

        @self.Constraint(
            self.flowsheet().time,
            self.zero_flow_components,
            doc="Components with no flow equation",
        )
        def return_zero_flow_comp(blk, t, i):
            return (
                blk.properties_out[t].conc_mass_comp[i]
                == 1e-10 * pyunits.kg / pyunits.m**3
            )

        @self.Constraint(
            self.flowsheet().time,
            doc="Equality alkalinity equation",
        )
        def return_Salk(blk, t):
            return (
                blk.properties_in[t].alkalinity
                == blk.properties_out[t].conc_mass_comp["S_IC"] / mw_c
            )

        iscale.set_scaling_factor(self.properties_out[0].flow_vol, 1e5)

    def initialize_build(
        self,
        state_args_in=None,
        state_args_out=None,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        This method calls the initialization method of the state blocks.

        Keyword Arguments:
            state_args_in : a dict of arguments to be passed to the inlet
                property package (to provide an initial state for
                initialization (see documentation of the specific
                property package) (default = None).
            state_args_out : a dict of arguments to be passed to the outlet
                property package (to provide an initial state for
                initialization (see documentation of the specific
                property package) (default = None).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)

        Returns:
            None
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")

        # Create solver
        opt = get_solver(solver, optarg)

        # ---------------------------------------------------------------------
        # Initialize state block
        flags = self.properties_in.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args_in,
            hold_state=True,
        )

        self.properties_out.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args_out,
        )

        if degrees_of_freedom(self) != 0:
            raise Exception(
                f"{self.name} degrees of freedom were not 0 at the beginning "
                f"of initialization. DoF = {degrees_of_freedom(self)}"
            )

        with idaeslog.solver_log(init_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)

        self.properties_in.release_state(flags=flags, outlvl=outlvl)

        init_log.info(f"Initialization Complete: {idaeslog.condition(res)}")

        if not check_optimal_termination(res):
            raise InitializationError(
                f"{self.name} failed to initialize successfully. Please check "
                f"the output logs for more information."
            )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pytest
import pyomo.environ as pyo

from pyomo.network import Port
from pyomo.util.check_units import assert_units_consistent
from idaes.core import (
    FlowsheetBlock,
    EnergyBalanceType,
    MaterialBalanceType,
    MomentumBalanceType,
    UnitModelCostingBlock,
)
from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    number_variables,
    number_total_constraints,
    number_unused_variables,
)
from idaes.core.util.scaling import (
    calculate_scaling_factors,
    unscaled_variables_generator,
    badly_scaled_var_generator,
)
from idaes.core.util.testing import initialization_tester
from watertap.property_models.multicomp_aq_sol_prop_pack import MCASParameterBlock
from watertap.unit_models.electrolyzer import Electrolyzer
from watertap.costing import WaterTAPCosting

__author__ = "Hunter Barber"

solver = get_solver()

# -----------------------------------------------------------------------------
class TestElectrolyzer:
    @pytest.fixture(scope="class")
    def chlor_alkali_elec(self):

        m = pyo.ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = MCASParameterBlock(
            solute_list=[
                "NA+",
                "CL-",
                "CL2-v",
                "H2-v",
                "OH-",
            ],
            mw_data={
                "H2O": 0.018015,
                "NA+": 0.022989,
                "CL-": 0.03545,
                "CL2-v": 0.0709,
                "H2-v": 0.002016,
                "OH-": 0.017007,
            },
            charge={"NA+": 1, "CL-": -1, "OH-": -1},
            ignore_neutral_charge=True,
        )

        m.fs.unit = Electrolyzer(
            property_package=m.fs.properties,
        )

        # fix property parameters
        m.fs.properties.dens_mass_const = 1200

        # feed specifications
        anolyte_blk = m.fs.unit.anolyte
        catholyte_blk = m.fs.unit.catholyte
        anolyte_blk.properties_in[0].pressure.fix(101325)
        anolyte_blk.properties_in[0].temperature.fix(273.15 + 90)
        anolyte_blk.properties_in[0].flow_mol_phase_comp["Liq", "H2O"].fix(5.551)
        anolyte_blk.properties_in[0].flow_mol_phase_comp["Liq", "NA+"].fix(0.3422)
        anolyte_blk.properties_in[0].flow_mol_phase_comp["Liq", "CL-"].fix(0.3422)
        anolyte_blk.properties_in[0].flow_mol_phase_comp["Liq", "CL2-v"].fix(0)
        anolyte_blk.properties_in[0].flow_mol_phase_comp["Liq", "H2-v"].fix(0)
        anolyte_blk.properties_in[0].flow_mol_phase_comp["Liq", "OH-"].fix(0)
        catholyte_blk.properties_in[0].pressure.fix(101325)
        catholyte_blk.properties_in[0].temperature.fix(273.15 + 90)
        catholyte_blk.properties_in[0].flow_mol_phase_comp["Liq", "H2O"].fix(5.551)
        catholyte_blk.properties_in[0].flow_mol_phase_comp["Liq", "NA+"].fix(1.288)
        catholyte_blk.properties_in[0].flow_mol_phase_comp["Liq", "CL-"].fix(0)
        catholyte_blk.properties_in[0].flow_mol_phase_comp["Liq", "CL2-v"].fix(0)
        catholyte_blk.properties_in[0].flow_mol_phase_comp["Liq", "H2-v"].fix(0)
        catholyte_blk.properties_in[0].flow_mol_phase_comp["Liq", "OH-"].fix(1.288)

        # touch properties
        anolyte_blk.properties_in[0].flow_vol_phase
        anolyte_blk.properties_in[0].conc_mass_phase_comp
        anolyte_blk.properties_in[0].conc_mol_phase_comp
        catholyte_blk.properties_in[0].flow_vol_phase
        catholyte_blk.properties_in[0].conc_mass_phase_comp
        catholyte_blk.properties_in[0].conc_mol_phase_comp
        anolyte_blk.properties_out[0].flow_vol_phase
        anolyte_blk.properties_out[0].conc_mass_phase_comp
        anolyte_blk.properties_out[0].conc_mol_phase_comp
        catholyte_blk.properties_out[0].flow_vol_phase
        catholyte_blk.properties_out[0].conc_mass_phase_comp
        catholyte_blk.properties_out[0].conc_mol_phase_comp

        return m

    @pytest.mark.unit
    def test_config(self, chlor_alkali_elec):

        m = chlor_alkali_elec
        u_config = m.fs.unit.config

        # check unit config arguments
        assert len(u_config) == 8
        assert not u_config.dynamic
        assert not u_config.has_holdup
        assert u_config.material_balance_type == MaterialBalanceType.useDefault
        assert u_config.energy_balance_type == EnergyBalanceType.none
        assert u_config.momentum_balance_type == MomentumBalanceType.pressureTotal

        # check properties
        assert u_config.property_package is m.fs.properties
        assert len(u_config.property_package.solute_set) == 5
        assert len(u_config.property_package.solvent_set) == 1

    @pytest.mark.unit
    def test_build(self, chlor_alkali_elec):

        m = chlor_alkali_elec

        # test units
        assert assert_units_consistent(m) is None

        # test ports
        port_lst = [
            "anolyte_inlet",
            "anolyte_outlet",
            "catholyte_inlet",
            "catholyte_outlet",
        ]
        for port_str in port_lst:
            port = getattr(m.fs.unit, port_str)
            assert len(port.vars) == 3
            assert isinstance(port, Port)

        # test statistics
        assert number_variables(m) == 212
        assert number_total_constraints(m) == 153
        assert number_unused_variables(m) == 15

    @pytest.mark.unit
    def test_dof(self, chlor_alkali_elec):

        m = chlor_alkali_elec

        # test initial degrees of freedom
        assert degrees_of_freedom(m) == 10

        # fix electrolysis reaction variables, TODO: transfer the following variables to generic importable blocks
        # membrane properties
        m.fs.unit.membrane_ion_transport_number["Liq", "NA+"].fix(1)
        # anode properties, Cl- --> 0.5 Cl2 + e-
        m.fs.unit.anode_electrochem_potential.fix(1.21)
        m.fs.unit.anode_stoich["Liq", "CL-"].fix(-1)
        m.fs.unit.anode_stoich["Liq", "CL2-v"].fix(0.5)
        # cathode properties, H20 + e- --> 0.5 H2 + OH-
        m.fs.unit.cathode_electrochem_potential.fix(-0.99)
        m.fs.unit.cathode_stoich["Liq", "H2O"].fix(-1)
        m.fs.unit.cathode_stoich["Liq", "H2-v"].fix(0.5)
        m.fs.unit.cathode_stoich["Liq", "OH-"].fix(1)

        # fix design and performance variables
        # membrane properties
        m.fs.unit.membrane_current_density.fix(4000)
        # anode properties
        m.fs.unit.anode_current_density.fix(3000)
        m.fs.unit.anode_overpotential.fix(0.1)  # assumed
        # cathode properties
        m.fs.unit.cathode_current_density.fix(3000)
        m.fs.unit.cathode_overpotential.fix(0.1)  # assumed
        # electrolyzer cell design
        m.fs.unit.current.fix(30000)
        # performance variables
        m.fs.unit.efficiency_current.fix(0.9)
        m.fs.unit.efficiency_voltage.fix(0.8)

        # test degrees of freedom satisfied
        assert degrees_of_freedom(m) == 0

    @pytest.mark.unit
    def test_init(self, chlor_alkali_elec):

        m = chlor_alkali_elec

        # set default scaling factors
        prop = m.fs.properties
        prop.set_default_scaling("flow_mol_phase_comp", 1, index=("Liq", "H2O"))
        prop.set_default_scaling("flow_mol_phase_comp", 1, index=("Liq", "NA+"))
        prop.set_default_scaling("flow_mol_phase_comp", 1, index=("Liq", "CL-"))
        prop.set_default_scaling("flow_mol_phase_comp", 1, index=("Liq", "CL2-v"))
        prop.set_default_scaling("flow_mol_phase_comp", 1, index=("Liq", "H2-v"))
        prop.set_default_scaling("flow_mol_phase_comp", 1, index=("Liq", "OH-"))

        calculate_scaling_factors(m)

        # check that all variables have scaling factors
        assert len(list(unscaled_variables_generator(m))) == 0

        # test initialization
        initialization_tester(m)

        # check variable scaling
        assert len(list(badly_scaled_var_generator(m, zero=1e-6))) == 0

    @pytest.mark.component
    def test_solve(self, chlor_alkali_elec):

        m = chlor_alkali_elec
        results = solver.solve(m)

        # check for optimal solution
        assert pyo.check_optimal_termination(results)

        # re-check variable scaling post solve
        assert len(list(badly_scaled_var_generator(m, zero=1e-6))) == 0

    @pytest.mark.component
    def test_solution(self, chlor_alkali_elec):

        m = chlor_alkali_elec

        # test report
        m.fs.unit.report()

        # check solution values
        assert pytest.approx(7.500, rel=1e-3) == pyo.value(m.fs.unit.membrane_area)
        assert pytest.approx(10.00, rel=1e-3) == pyo.value(m.fs.unit.anode_area)
        assert pytest.approx(10.00, rel=1e-3) == pyo.value(m.fs.unit.cathode_area)
        assert pytest.approx(2.750, rel=1e-3) == pyo.value(m.fs.unit.voltage_cell)
        assert pytest.approx(1.167e-5, rel=1e-3) == pyo.value(m.fs.unit.resistance)
        assert pytest.approx(82510, rel=1e-3) == pyo.value(m.fs.unit.power)
        assert pytest.approx(2.200, rel=1e-3) == pyo.value(m.fs.unit.voltage_reversible)
        assert pytest.approx(0.2798, rel=1e-3) == pyo.value(m.fs.unit.electron_flow)
        assert pytest.approx(0.7200, rel=1e-3) == pyo.value(m.fs.unit.efficiency_power)

        # check flow at outlet
        assert pytest.approx(0.1399, rel=1e-3) == pyo.value(
            m.fs.unit.anolyte.properties_out[0].flow_mol_phase_comp["Liq", "CL2-v"]
        )
        assert pytest.approx(1.568, rel=1e-3) == pyo.value(
            m.fs.unit.catholyte.properties_out[0].flow_mol_phase_comp["Liq", "NA+"]
        )
        assert pytest.approx(1.568, rel=1e-3) == pyo.value(
            m.fs.unit.catholyte.properties_out[0].flow_mol_phase_comp["Liq", "OH-"]
        )

        # check charge balance
        m.fs.unit.anolyte.properties_in[0].assert_electroneutrality(
            tee=True,
            tol=1e-5,
            solve=False,
        )
        m.fs.unit.anolyte.properties_out[0].assert_electroneutrality(
            tee=True,
            tol=1e-5,
            solve=False,
            defined_state=False,
        )
        m.fs.unit.catholyte.properties_in[0].assert_electroneutrality(
            tee=True,
            tol=1e-5,
            solve=False,
        )
        m.fs.unit.catholyte.properties_out[0].assert_electroneutrality(
            tee=True,
            tol=1e-5,
            solve=False,
            defined_state=False,
        )

    @pytest.mark.unit
    def test_costing_build(self, chlor_alkali_elec):

        m = chlor_alkali_elec

        # build costing model block
        m.fs.costing = WaterTAPCosting()
        m.fs.costing.base_currency = pyo.units.USD_2020
        m.fs.unit.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
        m.fs.costing.cost_process()

        # check general config after building costing block
        assert assert_units_consistent(m) is None
        assert degrees_of_freedom(m) == 0

    @pytest.mark.unit
    def test_costing_results(self, chlor_alkali_elec):

        m = chlor_alkali_elec

        # scale, initialize, and solve with costing
        calculate_scaling_factors(m)
        m.fs.unit.costing.initialize()
        results = solver.solve(m)

        # check for optimal solution
        assert pyo.check_optimal_termination(results)

        # check solution values
        assert pytest.approx(2.0 * 17930, rel=1e-3) == pyo.value(
            m.fs.unit.costing.capital_cost
        )
        assert pytest.approx(82.50, rel=1e-3) == pyo.value(
            m.fs.costing.aggregate_flow_electricity
        )
        assert pytest.approx(50040, rel=1e-3) == pyo.value(
            m.fs.costing.aggregate_flow_costs["electricity"]
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

# TODO: If needed for coverage, *build* and set operating conditions for,
#       but do not *solve*, the cross of all the options.

import glob
import math
import pytest
import os

import pandas as pd

from pyomo.environ import check_optimal_termination, value

from watertap.examples.flowsheets.lsrro.lsrro import (
    ACase,
    BCase,
    ABTradeoff,
    run_lsrro_case,
)

from .gha_divider import get_test_cases_subset

_this_file_path = os.path.dirname(os.path.abspath(__file__))

_input_headers = {
    "cin (kg/m3)": (float, "Cin"),
    "recovery (-)": (float, "water_recovery"),
    "num_stages": (int, "number_of_stages"),
    "A_case": (ACase, "A_case"),
    "B_case": (BCase, "B_case"),
    "AB_Tradeoff": (ABTradeoff, "AB_tradeoff"),
    "AB_gamma_factor": (float, "AB_gamma_factor"),
}

_results_headers = {
    "final brine concentration": "fs.disposal.properties[0].conc_mass_phase_comp[Liq, NaCl]",
    "final perm (ppm)": "fs.final_permeate_concentration",
    "Membrane area": "fs.total_membrane_area",
    "SEC": "fs.costing.specific_energy_consumption",
    "LCOW": "fs.costing.LCOW",
    "LCOW_feed": "fs.costing.LCOW_feed",
    "primary_pump_capex": "fs.costing.primary_pump_capex_lcow",
    "booster_pump_capex": "fs.costing.booster_pump_capex_lcow",
    "erd_capex": "fs.costing.erd_capex_lcow",
    "membrane_capex": "fs.costing.membrane_capex_lcow",
    "indirect_capex": "fs.costing.indirect_capex_lcow",
    "electricity": "fs.costing.electricity_lcow",
    "membrane_replacement": "fs.costing.membrane_replacement_lcow",
    "chem_lab_main": "fs.costing.chemical_labor_maintenance_lcow",
    "pumping_energy_agg_costs": "fs.costing.pumping_energy_aggregate_lcow",
    "membrane_agg_costs": "fs.costing.membrane_aggregate_lcow",
}

_csv_files = sorted(
    glob.glob(os.path.join(_this_file_path, "paper_analysis_baselines", "*.csv"))
)

_dfs = {os.path.basename(csv_file): pd.read_csv(csv_file) for csv_file in _csv_files}

_test_cases = [
    (csv_file, idx) for csv_file, df in _dfs.items() for idx in range(len(df))
]

# comment out this line if you want to run the entire baseline
_test_cases = get_test_cases_subset(_test_cases)


@pytest.mark.parametrize("csv_file, row_index", _test_cases)
@pytest.mark.component
def test_against_paper_analysis(csv_file, row_index):

    row = _dfs[csv_file].iloc[row_index]
    input_arguments = {
        argument: converter(row[property_name])
        for property_name, (converter, argument) in _input_headers.items()
    }
    number_of_stages = input_arguments["number_of_stages"]
    try:
        model, results = run_lsrro_case(
            **input_arguments,
            has_NaCl_solubility_limit=True,
            permeate_quality_limit=1000e-6,
            has_calculated_concentration_polarization=True,
            has_calculated_ro_pressure_drop=True,
            A_value=5 / 3.6e11,
            B_max=None,
            number_of_RO_finite_elements=10
        )
    except ValueError:
        results = None

    if results is not None and check_optimal_termination(results):
        for property_name, flowsheet_attribute in _results_headers.items():
            assert value(model.find_component(flowsheet_attribute)) == pytest.approx(
                float(row[property_name]),
                rel=1e-2,
            )
    else:
        for property_name in _results_headers:
            assert math.isnan(float(row[property_name]))



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Simple flowsheet interface API
"""

__author__ = "Dan Gunter"

# stdlib
from collections import namedtuple
from csv import reader, writer
from enum import Enum
from io import TextIOBase

try:
    from importlib.resources import files
except ImportError:
    from importlib_resources import files
import inspect
import logging
from pathlib import Path
import re
from typing import Any, Callable, List, Optional, Dict, Union, TypeVar
from types import ModuleType

try:
    from importlib import metadata
except ImportError:
    import importlib_metadata as metadata

# third-party
import idaes.logger as idaeslog
from idaes.core.util.model_statistics import degrees_of_freedom
from pydantic import BaseModel, validator, Field
import pyomo.environ as pyo

#: Forward-reference to a FlowsheetInterface type, used in
#: :meth:`FlowsheetInterface.find`
FSI = TypeVar("FSI", bound="FlowsheetInterface")


_log = idaeslog.getLogger(__name__)


class UnsupportedObjType(TypeError):
    def __init__(
        self,
        obj: Any,
        supported: Optional = None,
    ):
        msg = f"Object '{obj}' of type '{type(obj)}' is not supported."
        if supported is not None:
            msg += f"\nSupported: {supported}"
        super().__init__(msg)
        self.obj = obj
        self.supported = supported


class ModelExport(BaseModel):
    """A variable, expression, or parameter."""

    _SupportedObjType = Union[
        pyo.Var,
        pyo.Expression,
        pyo.Param,
    ]
    "Used for type hints and as a shorthand in error messages (i.e. not for runtime checks)"

    # TODO: if Optional[_SupportedObjType] is used for the `obj` type hint,
    # pydantic will run the runtime instance check which is not what we want
    # (as we want/need to use the pyomo is_xxx_type() methods instead)
    # so we're using Optional[object] unless we find a way to tell pydantic to skip this check
    # inputs
    obj: Optional[object] = Field(default=None, exclude=True)
    name: str = ""
    value: float = 0.0
    ui_units: object = Field(default=None, exclude=True)
    display_units: str = ""
    rounding: float = 0
    description: str = ""
    is_input: bool = True
    is_output: bool = True
    is_readonly: bool = None
    input_category: Optional[str]
    output_category: Optional[str]
    # computed
    obj_key: str = None
    fixed: bool = True
    lb: Union[None, float] = 0.0
    ub: Union[None, float] = 0.0
    num_samples: int = 2
    has_bounds: bool = True
    is_sweep: bool = False

    class Config:
        arbitrary_types_allowed = True

    @validator("obj", always=True, pre=True)
    @classmethod
    def ensure_obj_is_supported(cls, v):
        if v is not None:
            cls._ensure_supported_type(v)
        return v

    @classmethod
    def _ensure_supported_type(cls, obj: object):
        is_valid = (
            obj.is_variable_type()
            or obj.is_expression_type()
            or obj.is_parameter_type()
            # TODO: add support for numbers with pyo.numvalue.is_numeric_data()
        )
        if is_valid:
            return True
        raise UnsupportedObjType(obj, supported=cls._SupportedObjType)

    @classmethod
    def _get_supported_obj(
        cls, values: dict, field_name: str = "obj", allow_none: bool = False
    ):
        obj = values.get(field_name, None)
        if not allow_none and obj is None:
            raise TypeError(f"'{field_name}' is None but allow_none is False")
        cls._ensure_supported_type(obj)
        return obj

    # NOTE: IMPORTANT: all validators used to set a dynamic default value
    # should have the `always=True` option, or the validator won't be called
    # when the value for that field is not passed
    # (which is precisely when we need the default value)
    # additionally, `pre=True` should be given if the field can at any point
    # have a value that doesn't match its type annotation
    # (e.g. `None` for a strict (non-`Optional` `bool` field)

    # Get value from object
    @validator("value", always=True)
    @classmethod
    def validate_value(cls, v, values):
        if values.get("obj", None) is None:
            return v
        obj = cls._get_supported_obj(values, allow_none=False)
        return pyo.value(obj)

    # Derive display_units from ui_units
    @validator("display_units", always=True)
    @classmethod
    def validate_units(cls, v, values):
        if not v:
            u = values.get("ui_units", pyo.units.dimensionless)
            v = str(pyo.units.get_units(u))
        return v

    # set name dynamically from object
    @validator("name", always=True)
    @classmethod
    def validate_name(cls, v, values):
        if not v:
            obj = cls._get_supported_obj(values, allow_none=False)
            try:
                v = obj.name
            except AttributeError:
                pass
        return v

    @validator("is_readonly", always=True, pre=True)
    @classmethod
    def set_readonly_default(cls, v, values):
        if v is None:
            v = True
            obj = cls._get_supported_obj(values, allow_none=False)
            if obj.is_variable_type() or (obj.is_parameter_type() and obj.mutable):
                v = False
        return v

    @validator("obj_key", always=True, pre=True)
    @classmethod
    def set_obj_key_default(cls, v, values):
        if v is None:
            obj = cls._get_supported_obj(values, allow_none=False)
            v = str(obj)
        return v


class ModelOption(BaseModel):
    """An option for building/running the model."""

    name: str
    display_name: str = None
    description: str = None
    display_values: List[Any] = []
    values_allowed: List[Any] = []
    value: Any = None

    @validator("display_name", always=True)
    @classmethod
    def validate_display_name(cls, v, values):
        if v is None:
            v = values.get("name")
        return v

    @validator("description", always=True)
    @classmethod
    def validate_description(cls, v, values):
        if v is None:
            v = values.get("display_name")
        return v

    @validator("value")
    @classmethod
    def validate_value(cls, v, values):
        allowed = values.get("values_allowed", None)
        # allowed = list(allowed.keys())
        if v in allowed:
            return v
        else:
            raise ValueError(f"'value' ({v}) not in allowed values: {allowed}")


class FlowsheetExport(BaseModel):
    """A flowsheet and its contained exported model objects."""

    m: object = Field(default=None, exclude=True)
    obj: object = Field(default=None, exclude=True)
    name: str = ""
    description: str = ""
    model_objects: Dict[str, ModelExport] = {}
    version: int = 2
    requires_idaes_solver: bool = False
    dof: int = 0
    sweep_results: Union[None, dict] = {}
    build_options: Dict[str, ModelOption] = {}

    # set name dynamically from object
    @validator("name", always=True)
    @classmethod
    def validate_name(cls, v, values):
        if not v:
            try:
                v = values["obj"].name
            except (KeyError, AttributeError):
                pass
            if not v:
                v = "default"
        return v

    @validator("description", always=True)
    @classmethod
    def validate_description(cls, v, values):
        if not v:
            try:
                v = values["obj"].doc
            except (KeyError, AttributeError):
                v = f"{values['name']} flowsheet"
        return v

    def add(self, *args, data: Union[dict, ModelExport] = None, **kwargs) -> object:
        """Add a new variable (or other model object).

        There are a few different ways of invoking this function. Users will
        typically use this form::

            add(obj=<pyomo object>, name="My value name", ..etc..)

        where the keywords after `obj` match the non-computed names in :class:`ModelExport`.

        If these same name/value pairs are already in a dictionary, this form is more
        convenient::

            add(data=my_dict_of_name_value_pairs)

        If you have an existing ModelExport object, you can add it more directly with::

            add(my_object)
            # -- OR --
            add(data=my_object)


        Args:
            *args: If present, should be a single non-named argument, which is a
                 ModelExport object. Create by adding it.
            data: If present, create from this argument. If it's a dict, create from
                 its values just as from the kwargs. Otherwise, it should be a
                 ModelExport object, and create by adding it.
            kwargs: Name/value pairs to create a ModelExport object.
                    Accepted names and default values are in the ModelExport.

        Raises:
            KeyError: If the name of the Pyomo object is the same as an existing one,
                i.e. refuse to overwrite.
        """
        if len(args) > 1:
            raise ValueError(f"At most one non-keyword arg allowed. Got: {args}")
        if len(args) == 1:
            model_export = args[0]
        elif data is None:
            _log.debug(f"Create ModelExport from args: {kwargs}")
            model_export = ModelExport.parse_obj(kwargs)
        else:
            if isinstance(data, dict):
                model_export = ModelExport.parse_obj(data)
            else:
                model_export = data
        key = model_export.obj_key
        if key in self.model_objects:
            raise KeyError(
                f"Adding ModelExport object failed: duplicate key '{key}' (model_export={model_export})"
            )
        if _log.isEnabledFor(logging.DEBUG):  # skip except in debug mode
            _log.debug(
                f"Adding ModelExport object with key={key}: {model_export.dict()}"
            )
        self.model_objects[key] = model_export
        return model_export

    def from_csv(self, file: Union[str, Path], flowsheet):
        """Load multiple exports from the given CSV file.

        CSV file format rules:

            * Always use a header row. The names are case-insensitive, order is
              not important. The 'name', 'obj', and 'ui_units' columns are required.
            * Columns names should match the non-computed names in :class:`ModelExport`.
              See `.add()` for a list.
            * The object to export should be in a column named 'obj', prefixed with 'fs.'
            * For units, use Pyomo units module as 'units', e.g., 'mg/L' is `units.mg / units.L`

        For example::

            name,obj,description,ui_units,display_units,rounding,is_input,input_category,is_output,output_category
            Leach liquid feed rate,fs.leach_liquid_feed.flow_vol[0],Leach liquid feed volumetric flow rate,units.L/units.hour,L/h,2,TRUE,Liquid feed,FALSE,
            Leach liquid feed H,"fs.leach_liquid_feed.conc_mass_comp[0,'H']",Leach liquid feed hydrogen mass composition,units.mg/units.L,mg/L,3,TRUE,Liquid feed,FALSE,
            .......etc.......

        Args:
            file: Filename or path. If not an absolute path, start from the
                  directory of the caller's file.
            flowsheet: Flowsheet used to evaluate the exported objects.

        Returns:
            int: Number of exports added

        Raises:
            IOError: if input file doesn't exist
            ValueError: Invalid data in input file (error message will have details)
        """
        _log.debug(f"exports.add: from csv filename={file}")

        # compute path
        path = Path(file) if not isinstance(file, Path) else file
        if path.is_absolute():
            _log.debug(
                f"Reading CSV data for interface exports from " f"absolute path: {path}"
            )
            text = open(path, "r", encoding="utf-8").read()
        else:
            caller = inspect.getouterframes(inspect.currentframe())[1]
            caller_mod = inspect.getmodule(caller.frame).__name__
            if "." not in caller_mod:  # not in a package
                path = Path(caller.filename).parent / file
                text = open(path, "r", encoding="utf-8").read()
            else:
                caller_pkg = ".".join(caller_mod.split(".")[:-1])  # strip module
                _log.debug(
                    f"Reading CSV data for interface exports from: "
                    f"file={path}, module={caller_mod}, package={caller_pkg}"
                )
                try:
                    text = files(caller_pkg).joinpath(path).read_text()
                except Exception as err:
                    raise IOError(
                        f"Could not find CSV file '{path}' relative to file "
                        f"calling .add() in '{caller_mod}': {err}"
                    )

        # process CSV file
        rows = reader(re.split(r"\r?\n", text))
        # read and pre-process the header row
        raw_header = next(rows)
        header = [s.strip().lower() for s in raw_header]
        for req in "name", "obj", "ui_units":
            if req not in header:
                raise ValueError(
                    f"Bad CSV header: '{req}' column is required. data=" f"{header}"
                )
        num = 0
        for row in rows:
            if len(row) == 0:
                continue
            # build raw dict from values and header
            data = {k: v for k, v in zip(header, row)}
            # evaluate the object in the flowsheet
            try:
                data["obj"] = eval(data["obj"], {"fs": flowsheet})
            except Exception as err:
                raise ValueError(f"Cannot find object in flowsheet: {data['obj']}")
            # evaluate the units
            norm_units = data["ui_units"].strip()
            if norm_units in ("", "none", "-"):
                data["ui_units"] = pyo.units.dimensionless
            else:
                try:
                    data["ui_units"] = eval(norm_units, {"units": pyo.units})
                except Exception as err:
                    raise ValueError(f"Bad units '{norm_units}': {err}")
            # process boolean values (starting with 'is_')
            for k in data:
                if k.startswith("is_"):
                    v = data[k].lower()
                    if v == "true":
                        data[k] = True
                    elif v == "false":
                        data[k] = False
                    else:
                        raise ValueError(
                            f"Bad value '{data[k]}' "
                            f"for boolean argument '{k}': "
                            f"must be 'true' or 'false' "
                            f"(case-insensitive)"
                        )
            # add parsed export
            self.add(data=data)
            num += 1

        return num

    def to_csv(self, output: Union[TextIOBase, Path, str] = None) -> int:
        """Write wrapped objects as CSV.

        Args:
            output: Where to write CSV file. Can be a stream, path, or filename.

        Returns:
            Number of objects written into file.

        Raises:
            IOError: If path is given, and not writable
        """
        # open file for writing
        if isinstance(output, TextIOBase):
            output_file = output
        else:
            p = Path(output)
            output_file = p.open("w")

        # initialize
        csv_output_file = writer(output_file)

        # write header row
        obj = next(iter(self.model_objects.values()))
        values = ["obj", "ui_units"]
        col_idx_map = {}
        for i, field_name in enumerate(obj.dict()):
            # add to mapping of field name to column number
            col_idx_map[field_name] = i + 2
            # add column name
            values.append(field_name)
        csv_output_file.writerow(values)
        ncol = len(values)

        # write a row for each object
        num = 0
        for key, obj in self.model_objects.items():
            # initialize values list
            #   first 2 column values are object name and units
            obj_name = self._massage_object_name(key)
            units_str = self._massage_ui_units(str(obj.ui_units))
            values = [obj_name, units_str] + [""] * (ncol - 2)
            # add columns
            for field_name, field_value in obj.dict().items():
                values[col_idx_map[field_name]] = field_value
            # write row
            csv_output_file.writerow(values)
            num += 1

        return num

    @staticmethod
    def _massage_object_name(s):
        s1 = re.sub(r"\[([^]]*)\]", r"['\1']", s)  # quote everything in [brackets]
        s2 = re.sub(r"\['([0-9.]+)'\]", r"[\1]", s1)  # unquote [0.0] numbers
        return s2

    @staticmethod
    def _massage_ui_units(s):
        if s == "dimensionless":
            return ""
        return s

    def add_option(self, name: str, **kwargs) -> ModelOption:
        """Add an 'option' to the flowsheet that can be displayed and manipulated
        from the UI.

        Constructs a :class:`ModelOption` instance with provided args and adds it to
        the dict of options, keyed by its `name`.

        Args:
            name: Name of option (internal, for accessing the option)
            kwargs: Fields of :class:`ModelOption`
        """
        option = ModelOption(name=name, **kwargs)
        self.build_options[name] = option
        return option


class Actions(str, Enum):
    """Known actions that can be run.
    Actions that users should not run directly (unless they know what they are
    doing) are prefixed with an underscore.
    """

    build = "build"
    solve = "solve"
    export = "_export"
    diagram = "diagram"


class FlowsheetCategory(str, Enum):
    """Flowsheet Categories"""

    wastewater = "Wasterwater Recovery"
    desalination = "Desalination"


class FlowsheetInterface:
    """Interface between users, UI developers, and flowsheet models."""

    #: Function to look for in modules. See :meth:`find`.
    UI_HOOK = "export_to_ui"

    #: Type of item in list ``MissingObjectError.missing``.
    #: ``key`` is the unique key assigned to the variable,
    #: ``name`` is the variable name in the flowsheet
    MissingObject = namedtuple("MissingObject", "key name")

    class MissingObjectError(Exception):
        """Error returned if data in `load` refers to a variable not found in the
        target object.

        Use the `.missing` attribute of the error object to get the list  of
        MissingObjects.
        """

        def __init__(self, missing):
            num = len(missing)
            plural = "" if num == 1 else "s"
            things = [f"{m[1]}" for m in missing]
            super().__init__(
                f"{num} object{plural} not found in the model: {', '.join(things)}"
            )
            self.missing = [
                FlowsheetInterface.MissingObject(key=m[0], name=m[1]) for m in missing
            ]

    def __init__(
        self,
        fs: FlowsheetExport = None,
        do_build: Callable = None,
        do_export: Callable = None,
        do_solve: Callable = None,
        get_diagram: Callable = None,
        category: FlowsheetCategory = None,
        custom_do_param_sweep_kwargs: Dict = None,
        **kwargs,
    ):
        """Constructor.

        Args:
            fs: An existing wrapper to a flowsheet object. If this is not provided,
                then one will be constructed by passing the keyword arguments to
                the built-in pydantic ``parse_obj()`` method
                of :class:`FlowsheetExport`.
            do_build: Function to call to build the flowsheet. It should build the
                flowsheet model and return the `FlowsheetBlock`, which is typically
                the `fs` attribute of the model object. **Required**
            do_export: Function to call to export variables after the model is built.
                This will be called automatically by :meth:`build()`. **Required**
            do_solve: Function to solve the model. It should return the result
                that the solver itself returns. **Required**
            custom_do_param_sweep_kwargs: Option for setting up parallel solver using
                custom solve function.
            **kwargs: See `fs` arg. If the `fs` arg *is* provided, these are ignored.
        """
        if fs is None:
            self.fs_exp = FlowsheetExport.parse_obj(kwargs)
        else:
            self.fs_exp = fs
        self._actions = {}
        for arg, name in (
            (do_export, "export"),
            (do_build, "build"),
            (do_solve, "solve"),
        ):
            if arg:
                if not callable(arg):
                    raise TypeError(f"'do_{name}' argument must be callable")
                self.add_action(getattr(Actions, name), arg)
            else:
                raise ValueError(f"'do_{name}' argument is required")
        if callable(get_diagram):
            self.add_action("diagram", get_diagram)
        else:
            self.add_action("diagram", None)

        self._actions["custom_do_param_sweep_kwargs"] = custom_do_param_sweep_kwargs

    def build(self, **kwargs):
        """Build flowsheet

        Args:
            **kwargs: User-defined values

        Returns:
            None

        Raises:
            RuntimeError: If the build fails
        """
        try:
            self.run_action(Actions.build, **kwargs)
        except Exception as err:
            raise RuntimeError(f"Building flowsheet: {err}") from err
        return

    def solve(self, **kwargs):
        """Solve flowsheet.

        Args:
            **kwargs: User-defined values

        Returns:
            Return value of the underlying solve function

        Raises:
            RuntimeError: if the solver did not terminate in an optimal solution
        """
        try:
            result = self.run_action(Actions.solve, **kwargs)
        except Exception as err:
            raise RuntimeError(f"Solving flowsheet: {err}") from err
        return result

    def get_diagram(self, **kwargs):
        """Return diagram image name.

        Args:
            **kwargs: User-defined values

        Returns:
            Return image file name if get_diagram function is callable. Otherwise, return none
        """
        if self.get_action(Actions.diagram) is not None:
            return self.run_action(Actions.diagram, **kwargs)
        else:
            return None

    def dict(self) -> Dict:
        """Serialize.

        Returns:
            Serialized contained FlowsheetExport object
        """
        return self.fs_exp.dict(exclude={"obj"})

    def load(self, data: Dict):
        """Load values from the data into corresponding variables in this
        instance's FlowsheetObject.

        Args:
            data: The input flowsheet (probably deserialized from JSON)
        """
        u = pyo.units
        fs = FlowsheetExport.parse_obj(data)  # new instance from data
        # Set the value for each input variable
        missing = []
        # 'src' is the data source and 'dst' is this flowsheet (destination)
        for key, src in fs.model_objects.items():
            # get corresponding exported variable
            try:
                dst = self.fs_exp.model_objects[key]
            except KeyError:
                missing.append((key, src.name))
                continue
            # set value in this flowsheet
            ui_units = dst.ui_units
            if dst.is_input and not dst.is_readonly:
                # only update if value has changed
                if dst.value != src.value:
                    # print(f'changing value for {key} from {dst.value} to {src.value}')
                    # create a Var so Pyomo can do the unit conversion for us
                    tmp = pyo.Var(initialize=src.value, units=ui_units)
                    tmp.construct()
                    # Convert units when setting value in the model
                    new_val = pyo.value(u.convert(tmp, to_units=u.get_units(dst.obj)))
                    # print(f'changing value for {key} from {dst.value} to {new_val}')
                    dst.obj.set_value(new_val)
                    # Don't convert units when setting the exported value
                    dst.value = src.value

                if dst.obj.fixed != src.fixed:
                    # print(f'changing fixed for {key} from {dst.obj.fixed} to {src.fixed}')
                    if src.fixed:
                        dst.obj.fix()
                    else:
                        dst.obj.unfix()
                    dst.fixed = src.fixed

                if dst.is_sweep != src.is_sweep:
                    dst.is_sweep = src.is_sweep

                if dst.num_samples != src.num_samples:
                    dst.num_samples = src.num_samples

                # update bounds
                if dst.lb != src.lb:
                    # print(f'changing lb for {key} from {dst.lb} to {src.lb}')
                    if src.lb is None or src.lb == "":
                        dst.obj.setlb(None)
                        dst.lb = None
                    else:
                        tmp = pyo.Var(initialize=src.lb, units=ui_units)
                        tmp.construct()
                        new_lb = pyo.value(
                            u.convert(tmp, to_units=u.get_units(dst.obj))
                        )
                        dst.obj.setlb(new_lb)
                        dst.lb = src.lb
                if dst.ub != src.ub:
                    # print(f'changing ub for {key} from {dst.ub} to {src.ub}')
                    if src.ub is None or src.ub == "":
                        dst.obj.setub(None)
                        dst.ub = None
                    else:
                        tmp = pyo.Var(initialize=src.ub, units=ui_units)
                        tmp.construct()
                        new_ub = pyo.value(
                            u.convert(tmp, to_units=u.get_units(dst.obj))
                        )
                        # print(f'changing ub for {key} from {dst.obj.ub} to {new_ub}')
                        dst.obj.setub(new_ub)
                        dst.ub = src.ub

        # update degrees of freedom (dof)
        self.fs_exp.dof = degrees_of_freedom(self.fs_exp.obj)
        if missing:
            raise self.MissingObjectError(missing)

    def select_option(self, option_name: str, new_option: str):
        """Update flowsheet with selected option.

        Args:
            data: The input flowsheet
            option_name: Name of selected option

        Returns:
            None
        """

        # fs = FlowsheetExport.parse_obj(data)  # new instance from data
        self.fs_exp.build_options[option_name].value = new_option

        # # get function name from model options
        # func_name = self.fs_exp.build_options[option_name].values_allowed[new_option]

        # # add functino name as new build function
        # self.add_action("build", func_name)

    def add_action(self, action_name: str, action_func: Callable):
        """Add an action for the flowsheet.

        Args:
            action_name: Name of the action to take (see :class:`Actions`)
            action_func: Function to call for the action

        Returns:
            None
        """

        # print(f'ADDING ACTION: {action_name}')
        # print(action_func)
        def action_wrapper(**kwargs):
            if action_name == Actions.build:
                # set new model object from return value of build action
                action_result = action_func(**kwargs)
                if action_result is None:
                    raise RuntimeError(
                        f"Flowsheet `{Actions.build}` action failed. "
                        f"See logs for details."
                    )
                self.fs_exp.obj = action_result.fs
                self.fs_exp.m = action_result
                # [re-]create exports (new model object)
                if Actions.export not in self._actions:
                    raise KeyError(
                        "Error in 'build' action: no export action defined. "
                        "Add `do_export=<function>` to FlowsheetInterface "
                        "constructor or call `add_action(Actions.export, <function>)` "
                        "on FlowsheetInterface instance."
                    )
                # clear model_objects dict, since duplicates not allowed
                self.fs_exp.model_objects.clear()
                # use get_action() since run_action() will refuse to call it directly
                self.get_action(Actions.export)(
                    exports=self.fs_exp, build_options=self.fs_exp.build_options
                )
                result = None
            elif action_name == Actions.diagram:
                self._actions[action_name] = action_func
                return
            elif self.fs_exp.obj is None:
                raise RuntimeError(
                    f"Cannot run any flowsheet action (except "
                    f"'{Actions.build}') before flowsheet is built"
                )
            else:
                result = action_func(flowsheet=self.fs_exp.obj, **kwargs)
                # Issue 755: Report optimization errors
                if action_name == Actions.solve:
                    _log.debug(f"Solve result: {result}")
                    if result is None:
                        raise RuntimeError("Solver did not return a result")
                    if not pyo.check_optimal_termination(result):
                        raise RuntimeError(f"Solve failed: {result}")
            # Sync model with exported values
            if action_name in (Actions.build, Actions.solve):
                self.export_values()
            return result

        self._actions[action_name] = action_wrapper

    def get_action(self, name: str) -> Union[Callable, None]:
        """Get the function for an ``add()``-ed action.

        Args:
            name: Name of the action (see :class:`Actions`)

        Returns:
            Function for this action

        Raises:
            KeyError, if no such action is defined
        """
        return self._actions[name]

    def run_action(self, name, **kwargs):
        """Run the named action."""
        func = self.get_action(name)
        if name.startswith("_"):
            raise ValueError(
                f"Refusing to call '{name}' action directly since its "
                f"name begins with an underscore"
            )
        return func(**kwargs)

    def export_values(self):
        """Copy current values in underlying Pyomo model into exported model.

        Side-effects:
            Attribute ``fs_exp`` is modified.
        """
        _log.info("Exporting values from flowsheet model to UI")
        u = pyo.units
        self.fs_exp.dof = degrees_of_freedom(self.fs_exp.obj)
        for key, mo in self.fs_exp.model_objects.items():
            mo.value = pyo.value(u.convert(mo.obj, to_units=mo.ui_units))
            # print(f'{key} is being set to: {mo.value}')
            if hasattr(mo.obj, "bounds"):
                # print(f'{key} is being set to: {mo.value} from {mo.obj.value}')
                if mo.obj.ub is None:
                    mo.ub = mo.obj.ub
                else:
                    tmp = pyo.Var(initialize=mo.obj.ub, units=u.get_units(mo.obj))
                    tmp.construct()
                    mo.ub = pyo.value(u.convert(tmp, to_units=mo.ui_units))
                if mo.obj.lb is None:
                    mo.lb = mo.obj.lb
                else:
                    tmp = pyo.Var(initialize=mo.obj.lb, units=u.get_units(mo.obj))
                    tmp.construct()
                    mo.lb = pyo.value(u.convert(tmp, to_units=mo.ui_units))
                mo.fixed = mo.obj.fixed
            else:
                mo.has_bounds = False

    @classmethod
    def from_installed_packages(
        cls, group_name: str = "watertap.flowsheets"
    ) -> Dict[str, "FlowsheetInterface"]:
        """Get all flowsheet interfaces defined as entry points within the Python packages installed in the environment.

        This uses the :func:`importlib.metadata.entry_points` function to fetch the
        list of flowsheets declared as part of a Python package distribution's `entry points <https://docs.python.org/3/library/importlib.metadata.html#entry-points>`_
        under the group ``group_name``.

        To set up a flowsheet interface for discovery, locate your Python package distribution's file (normally
        :file:`setup.py`, :file:`pyproject.toml`, or equivalent) and add an entry in the ``entry_points`` section.

        For example, to add a flowsheet defined in :file:`watertap/examples/flowsheets/my_flowsheet.py`
        so that it can be discovered with the name ``my_flowsheet`` wherever the ``watertap`` package is installed,
        the following should be added to WaterTAP's :file:`setup.py`::

           setup(
               name="watertap",
               # other setup() sections
               entry_points={
                   "watertap.flowsheets": [
                        # other flowsheet entry points
                        "my_flowsheet = watertap.examples.flowsheets.my_flowsheet",
                   ]
               }
           )

        Args:
            group_name: The entry_points group from which the flowsheet interface modules will be populated.

        Returns:
            Mapping with keys the module names and values FlowsheetInterface objects
        """
        eps = metadata.entry_points()
        try:
            # this happens for Python 3.7 (via importlib_metadata) and Python 3.10+
            entry_points = list(eps.select(group=group_name))
        except AttributeError:
            # this will happen on Python 3.8 and 3.9, where entry_points() has dict-like group selection
            entry_points = list(eps[group_name])

        if not entry_points:
            _log.error(f"No interfaces found for entry points group: {group_name}")
            return {}

        interfaces = {}
        _log.debug(f"Loading {len(entry_points)} entry points")
        for ep in entry_points:
            _log.debug(f"ep = {ep}")
            module_name = ep.value
            try:
                module = ep.load()
            except ImportError as err:
                _log.error(f"Cannot import module '{module_name}': {err}")
                continue
            interface = cls.from_module(module)
            if interface:
                interfaces[module_name] = interface

        return interfaces

    @classmethod
    def from_module(
        cls, module: Union[str, ModuleType]
    ) -> Optional["FlowsheetInterface"]:
        """Get a a flowsheet interface for module.

        Args:
            module: The module

        Returns:
            A flowsheet interface or None if it failed
        """
        if not isinstance(module, ModuleType):
            module = importlib.import_module(module)

        # Get function that creates the FlowsheetInterface
        func = getattr(module, cls.UI_HOOK, None)
        if func is None:
            _log.warning(
                f"Interface for module '{module}' is missing UI hook function: "
                f"{cls.UI_HOOK}()"
            )
            return None
        # Call the function that creates the FlowsheetInterface
        try:
            interface = func()
        except Exception as err:
            _log.error(
                f"Cannot get FlowsheetInterface object for module '{module}': {err}"
            )
            return None
        # Return created FlowsheetInterface
        return interface



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
#################################################################################

import pytest
from pathlib import Path
from os.path import join
from pandas import read_csv


@pytest.fixture
def get_test_file():
    test_file = read_csv(join(Path(__file__).parents[1], "periodic_table.csv"))
    return test_file


@pytest.mark.unit
def test_periodic_table_headers(get_test_file):
    test_headers = ["AtomicMass", "Symbol"]
    assert all(header in get_test_file.columns for header in test_headers)
    size = {"cols": 28, "rows": 118}
    assert len(get_test_file.columns) == size["cols"]
    assert len(get_test_file.values) == size["rows"]



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

"""
This module contains a zero-order representation of a microbial battery water
treatment unit.
"""

import pyomo.environ as pyo
from pyomo.environ import Var, units as pyunits
from idaes.core import declare_process_block_class
from watertap.core import build_sido_reactive, constant_intensity, ZeroOrderBaseData


# Some more information about this module
__author__ = "Travis Arnold"


@declare_process_block_class("MicrobialBatteryZO")
class MicrobialBatteryData(ZeroOrderBaseData):
    """
    Zero-Order model for a microbial battery water treatment unit.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "microbial_battery"

        build_sido_reactive(self)
        constant_intensity(self)

        # Create hydraulic retention time variable
        self.HRT = Var(
            units=pyunits.hr,
            bounds=(0, None),
            doc="Hydraulic retention time of water treatment unit",
        )
        self._perf_var_dict["Hydraulic Retention Time"] = self.HRT
        self._fixed_perf_vars.append(self.HRT)

        # Create reactor volume variable
        self.reactor_volume = Var(
            units=pyunits.m**3,
            bounds=(0, None),
            doc="Volume of water treatment unit",
        )
        self._perf_var_dict["Reactor Volume"] = self.reactor_volume

        @self.Constraint(self.flowsheet().time, doc="Constraint for reactor volume.")
        def reactor_volume_rule(b, t):
            return b.reactor_volume == (
                pyunits.convert(
                    b.HRT * b.properties_in[t].flow_vol, to_units=pyunits.m**3
                )
            )

    @property
    def default_costing_method(self):
        return self.cost_microbial_battery

    @staticmethod
    def cost_microbial_battery(blk):
        """
        General method for costing microbial battery treatment unit.
        """
        t0 = blk.flowsheet().time.first()

        # Get parameter dict from database
        parameter_dict = blk.unit_model.config.database.get_unit_operation_parameters(
            blk.unit_model._tech_type, subtype=blk.unit_model.config.process_subtype
        )

        # Get costing parameter sub-block for this technology
        sizing_cost = blk.unit_model._get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            ["sizing_cost"],
        )

        # Add cost variable and constraint
        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation",
        )

        expr = pyo.units.convert(
            blk.unit_model.properties_in[t0].flow_vol * sizing_cost,
            to_units=blk.config.flowsheet_costing_block.base_currency,
        )

        # Determine if a costing factor is required
        blk.costing_package.add_cost_factor(
            blk, parameter_dict["capital_cost"]["cost_factor"]
        )

        blk.capital_cost_constraint = pyo.Constraint(
            expr=blk.capital_cost == blk.cost_factor * expr
        )

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity"
        )
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.properties_in[t0].flow_mass_comp["filtration_media"],
            "filtration_media",
        )
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.properties_byproduct[t0].flow_mass_comp["filtration_media"],
            "filtration_media_disposal",
        )
